#!/usr/bin/env python
# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2012 University of North Carolina at Chapel Hill, Luke Campagnola
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module contains utils relative to qt Signal
"""

from silx.gui import qt
import weakref
from time import time
from silx.gui.utils import concurrent

__all__ = ['SignalProxy']
__authors__ = ['L. Campagnola', 'M. Liberty']
__license__ = "MIT"


class SignalProxy(qt.QObject):
    """
    This peace of code come from pyqtgraph
    Object which collects rapid-fire signals and condenses them
    into a single signal or a rate-limited stream of signals.
    Used, for example, to prevent a SpinBox from generating multiple
    signals when the mouse wheel is rolled over it.

    Emits sigDelayed after input signals have stopped for a certain period of time.
    """

    sigDelayed = qt.Signal(object)

    def __init__(self, signal, delay=0.3, rateLimit=0, slot=None):
        """Initialization arguments:
        signal - a bound Signal or pyqtSignal instance
        delay - Time (in seconds) to wait for signals to stop before emitting (default 0.3s)
        slot - Optional function to connect sigDelayed to.
        rateLimit - (signals/sec) if greater than 0, this allows signals to stream out at a
                    steady rate while they are being received.
        """

        qt.QObject.__init__(self)
        signal.connect(self.signalReceived)
        self.signal = signal
        self.delay = delay
        self.rateLimit = rateLimit
        self.args = None
        self.timer = qt.QTimer()
        self.timer.timeout.connect(self.flush)
        self.blockSignal = False
        self.slot = weakref.ref(slot)
        self.lastFlushTime = None
        if slot is not None:
            self.sigDelayed.connect(slot)

    def setDelay(self, delay):
        self.delay = delay

    def signalReceived(self, *args):
        """Received signal. Cancel previous timer and store args to be forwarded later."""
        if self.blockSignal:
            return
        self.args = args
        if self.rateLimit == 0:
            concurrent.submitToQtMainThread(self.timer.stop)
            concurrent.submitToQtMainThread(self.timer.start, (self.delay * 1000) + 1)
        else:
            now = time()
            if self.lastFlushTime is None:
                leakTime = 0
            else:
                lastFlush = self.lastFlushTime
                leakTime = max(0, (lastFlush + (1.0 / self.rateLimit)) - now)

            concurrent.submitToQtMainThread(self.timer.stop)
            concurrent.submitToQtMainThread(self.timer.start, (min(leakTime, self.delay) * 1000) + 1)
            # self.timer.stop()
            # self.timer.start((min(leakTime, self.delay) * 1000) + 1)

    def flush(self):
        """If there is a signal queued up, send it now."""
        if self.args is None or self.blockSignal:
            return False
        args, self.args = self.args, None
        concurrent.submitToQtMainThread(self.timer.stop)
        self.lastFlushTime = time()
        # self.emit(self.signal, *self.args)
        concurrent.submitToQtMainThread(self.sigDelayed.emit, args)
        # self.sigDelayed.emit(args)
        return True

    def disconnect(self):
        self.blockSignal = True
        try:
            self.signal.disconnect(self.signalReceived)
        except:
            pass
        try:
            self.sigDelayed.disconnect(self.slot)
        except:
            pass


if __name__ == '__main__':
    app = qt.QApplication([])
    win = qt.QMainWindow()
    spin = qt.QSpinBox()
    win.setCentralWidget(spin)
    win.show()


    def fn(*args):
        print("Raw signal:", args)


    def fn2(*args):
        print("Delayed signal:", args)


    spin.valueChanged.connect(fn)
    # proxy = proxyConnect(spin, QtCore.SIGNAL('valueChanged(int)'), fn)
    proxy = SignalProxy(spin.valueChanged, delay=0.5, slot=fn2)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2004-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
:mod:`silx.gui.plot.actions.mode` provides a set of QAction relative to mouse
mode of a :class:`.PlotWidget`.

The following QAction are available:

- :class:`ZoomModeAction`
- :class:`PanModeAction`
"""

from __future__ import division

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "16/08/2017"

from . import PlotAction
import logging

_logger = logging.getLogger(__name__)


class ZoomModeAction(PlotAction):
    """QAction controlling the zoom mode of a :class:`.PlotWidget`.

    :param plot: :class:`.PlotWidget` instance on which to operate
    :param parent: See :class:`QAction`
    """

    def __init__(self, plot, parent=None):
        super(ZoomModeAction, self).__init__(
            plot, icon='zoom', text='Zoom mode',
            tooltip='Zoom in or out',
            triggered=self._actionTriggered,
            checkable=True, parent=parent)
        # Listen to mode change
        self.plot.sigInteractiveModeChanged.connect(self._modeChanged)
        # Init the state
        self._modeChanged(None)

    def _modeChanged(self, source):
        modeDict = self.plot.getInteractiveMode()
        old = self.blockSignals(True)
        self.setChecked(modeDict["mode"] == "zoom")
        self.blockSignals(old)

    def _actionTriggered(self, checked=False):
        plot = self.plot
        if plot is not None:
            plot.setInteractiveMode('zoom', source=self)


class PanModeAction(PlotAction):
    """QAction controlling the pan mode of a :class:`.PlotWidget`.

    :param plot: :class:`.PlotWidget` instance on which to operate
    :param parent: See :class:`QAction`
    """

    def __init__(self, plot, parent=None):
        super(PanModeAction, self).__init__(
            plot, icon='pan', text='Pan mode',
            tooltip='Pan the view',
            triggered=self._actionTriggered,
            checkable=True, parent=parent)
        # Listen to mode change
        self.plot.sigInteractiveModeChanged.connect(self._modeChanged)
        # Init the state
        self._modeChanged(None)

    def _modeChanged(self, source):
        modeDict = self.plot.getInteractiveMode()
        old = self.blockSignals(True)
        self.setChecked(modeDict["mode"] == "pan")
        self.blockSignals(old)

    def _actionTriggered(self, checked=False):
        plot = self.plot
        if plot is not None:
            plot.setInteractiveMode('pan', source=self)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


import inspect
import types
import logging

import six


debug_logger = logging.getLogger("silx.DEBUG")

_indent = 0


def log_method(func, class_name=None):
    """Decorator to inject a warning log before an after any function/method.

    .. code-block:: python

        @log_method
        def foo():
            return None

    :param callable func: The function to patch
    :param str class_name: In case a method, provide the class name
    """
    def wrapper(*args, **kwargs):
        global _indent

        indent = "  " * _indent
        func_name = func.func_name if six.PY2 else func.__name__
        if class_name is not None:
            name = "%s.%s" % (class_name, func_name)
        else:
            name = "%s" % (func_name)

        debug_logger.warning("%s%s" % (indent, name))
        _indent += 1
        result = func(*args, **kwargs)
        _indent -= 1
        debug_logger.warning("%sreturn  (%s)" % (indent, name))
        return result
    return wrapper


def log_all_methods(base_class):
    """Decorator to inject a warning log before an after any method provided by
    a class.

    .. code-block:: python

        @log_all_methods
        class Foo(object):

            def a(self):
                return None

            def b(self):
                return self.a()

    Here is the output when calling the `b` method.

    .. code-block::

        WARNING:silx.DEBUG:_Foobar.b
        WARNING:silx.DEBUG:  _Foobar.a
        WARNING:silx.DEBUG:  return  (_Foobar.a)
        WARNING:silx.DEBUG:return  (_Foobar.b)

    :param class base_class: The class to patch
    """
    methodTypes = (types.MethodType, types.FunctionType, types.BuiltinFunctionType, types.BuiltinMethodType)
    for name, func in inspect.getmembers(base_class):
        if isinstance(func, methodTypes):
            if func.__name__ not in ["__subclasshook__", "__new__"]:
                # patching __new__ in Python2 break the object, then we skip it
                setattr(base_class, name, log_method(func, base_class.__name__))

    return base_class



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/silx-kit/silx
#
#    Copyright (C) 2013-2018  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

"""
Contains a class for creating a plan, allocating arrays, compiling kernels and
other things like that...
to calculate SIFT keypoints and descriptors.


This code implements the SIFT algorithm
The SIFT algorithm belongs to the University of British Columbia. It is
protected by patent US6711293. If you are on a country where this pattent
applies (like the USA), please check if you are allowed to use it. The
University of British Columbia does not require a license for its use for
non-commercial research applications.


This algorithm is patented: U.S. Patent 6,711,293:
"Method and apparatus for identifying scale invariant features in an image and use of same for locating an object in an image",
David Lowe's patent for the SIFT algorithm,  Mar. 8, 1999. 
It is due to expire in March 2019. 
"""

from __future__ import division, print_function, with_statement

__authors__ = ["Jérôme Kieffer", "Pierre Paleo"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "10/11/2020"
__status__ = "production"

import os
import time
import math
import logging
import gc
import numpy
from collections import OrderedDict
from .param import par
from silx.opencl import ocl, pyopencl, kernel_workgroup_size
from silx.opencl.utils import get_opencl_code, nextpower
from ..processing import OpenclProcessing, BufferDescription
from .utils import calc_size, kernel_size
logger = logging.getLogger(__name__)


class SiftPlan(OpenclProcessing):
    """This class implements a way to calculate SIFT keypoints.


    How to calculate a set of SIFT keypoint on an image::

        siftp = sift.SiftPlan(img.shape,img.dtype,devicetype="GPU")
        kp = siftp.keypoints(img)

    kp is a nx132 array. the second dimension is composed of x,y, scale and angle as well as 128 floats describing the keypoint

    This SIFT algorithm is patented: U.S. Patent 6,711,293:
    "Method and apparatus for identifying scale invariant features in an image and use of same for locating an object in an image",
    """
    kernels_max_wg_size = {"orientation_cpu": 1,
                           "orientation_gpu": 128,
                           "descriptor_gpu1": (8, 4, 4),
                           "descriptor_gpu2": (8, 8, 8),
                           "descriptor_cpu": (1,),
                           }

    converter = {numpy.dtype(numpy.uint8): "u8_to_float",
                 numpy.dtype(numpy.uint16): "u16_to_float",
                 numpy.dtype(numpy.uint32): "u32_to_float",
                 numpy.dtype(numpy.uint64): "u64_to_float",
                 numpy.dtype(numpy.int32): "s32_to_float",
                 numpy.dtype(numpy.int64): "s64_to_float",
                 # numpy.dtype(numpy.float64): "double_to_float",
                 }

    sigmaRatio = 2.0 ** (1.0 / par.Scales)
    PIX_PER_KP = 10  # pre_allocate buffers for keypoints
    dtype_kp = numpy.dtype([('x', numpy.float32),
                            ('y', numpy.float32),
                            ('scale', numpy.float32),
                            ('angle', numpy.float32),
                            ('desc', (numpy.uint8, 128))
                            ])

    def __init__(self, shape=None, dtype=None, template=None,
                 PIX_PER_KP=None, init_sigma=None,
                 ctx=None, devicetype="all", platformid=None, deviceid=None,
                 block_size=None, memory=None, profile=False):
        """
        Constructor of the class

        :param shape: shape of the input image
        :param dtype: data type of the input image
        :param devicetype: can be 'CPU' or 'GPU'
        :param template: extract shape and dtype from an image
        :param profile: collect timing info
        :param device: 2-tuple of integers
        :param PIX_PER_KP: number of keypoint pre-allocated: 1 for 10 pixel
        :param block_size: set to 1 under macosX on CPU
        :param context: provide an external context
        :param init_sigma: blurring width, you should have good reasons to modify 
                            the 1.6 default value...
        """
        self.kernels_max_wg_size = self.__class__.kernels_max_wg_size.copy()
        if template is not None:
            self.shape = template.shape
            self.dtype = template.dtype
        else:
            self.shape = shape
            self.dtype = numpy.dtype(dtype)
        if len(self.shape) == 3:
            self.RGB = True
            self.shape = self.shape[:2]
        elif len(self.shape) == 2:
            self.RGB = False
        else:
            raise RuntimeError("Unable to process image of shape %s" % (tuple(self.shape,)))
        if PIX_PER_KP:
            self.PIX_PER_KP = int(PIX_PER_KP)

        self.kpsize = None

        if init_sigma is None:
            init_sigma = par.InitSigma
        # no test on the values, just make sure it is a float
        self._init_sigma = float(init_sigma)
        memory = self._calc_memory(block_size)
        OpenclProcessing.__init__(self, ctx=ctx, devicetype=devicetype,
                                  platformid=platformid, deviceid=deviceid,
                                  block_size=block_size, profile=profile,
                                  memory=memory)
        # TODO WORKGROUP ....
        self.kernels_wg = {}
        self.scales = []  # in XY order
        self.procsize = []  # same as  procsize but with dimension in (X,Y) not (slow, fast)
        self.wgsize = []

        self.octave_max = None
        self.red_size = None
        self._calc_scales()
        self.LOW_END = 0
        self._calc_workgroups()
        self.compile_kernels()
        self._allocate_buffers()
        self.cnt = numpy.empty(1, dtype=numpy.int32)
        if "CPU" in self.device.type:
            self.USE_CPU = True
        else:
            self.USE_CPU = False

    def _calc_scales(self):
        """
        Nota scales are in XY order
        """
        shape = self.shape[-1::-1]
        self.scales = [tuple(numpy.int32(i) for i in shape)]
        min_size = 2 * par.BorderDist + 2
        while min(shape) > min_size:
            shape = tuple(numpy.int32(i // 2) for i in shape)
            self.scales.append(shape)
        self.scales.pop()
        self.octave_max = len(self.scales)

    def _calc_memory(self, block_size=None):
        """
        Estimates the memory footprint of all buffer to ensure it fits on the device
        """
        block_size = int(block_size) if block_size else 4096  # upper limit

        # Just the context + kernel takes about 75MB on the GPU
        memory = 75 * 2 ** 20
        size_of_float = numpy.dtype(numpy.float32).itemsize
        size_of_input = numpy.dtype(self.dtype).itemsize
        # raw images:
        size = self.shape[0] * self.shape[1]
        memory += size * size_of_input  # initial_image (no raw_float)
        if self.RGB:
            memory += 2 * size * (size_of_input)  # one of three was already counted
        nr_blur = par.Scales + 3  # 3 blurs and 2 tmp
        nr_dogs = par.Scales + 2
        memory += size * (nr_blur + nr_dogs) * size_of_float

        self.kpsize = int(self.shape[0] * self.shape[1] // self.PIX_PER_KP)  # Is the number of kp independant of the octave ? int64 causes problems with pyopencl
        memory += self.kpsize * size_of_float * 4 * 2  # those are array of float4 to register keypoints, we need two of them
        memory += self.kpsize * 128  # stores the descriptors: 128 unsigned chars
        memory += 4  # keypoint index Counter
        wg_float = min(block_size, numpy.sqrt(self.shape[0] * self.shape[1]))
        self.red_size = nextpower(wg_float)
        memory += 4 * 2 * self.red_size  # temporary storage for reduction

        ########################################################################
        # Calculate space for gaussian kernels
        ########################################################################
        curSigma = 1.0 if par.DoubleImSize else 0.5
        if self._init_sigma > curSigma:
            sigma = math.sqrt(self._init_sigma ** 2 - curSigma ** 2)
            size = kernel_size(sigma, True)
            logger.debug("pre-Allocating %s float for init blur" % size)
            memory += size * size_of_float
        prevSigma = self._init_sigma
        for _ in range(par.Scales + 2):
            increase = prevSigma * math.sqrt(self.sigmaRatio ** 2 - 1.0)
            size = kernel_size(increase, True)
            logger.debug("pre-Allocating %s float for blur sigma: %s" % (size, increase))
            memory += size * size_of_float
            prevSigma *= self.sigmaRatio
        # self.memory = memory
        return memory

    def _allocate_buffers(self):
        """
        All buffers are allocated here
        """
        shape = self.shape
        buffers = [BufferDescription("min", 1, numpy.float32, None),
                   BufferDescription("max", 1, numpy.float32, None),
                   BufferDescription("255", 1, numpy.float32, None),
                   BufferDescription("cnt", 1, numpy.int32, None),
                   BufferDescription("Kp_1", (self.kpsize, 4), numpy.float32, None),
                   BufferDescription("Kp_2", (self.kpsize, 4), numpy.float32, None),
                   BufferDescription("descr", (self.kpsize, 128), numpy.uint8, None),
                   BufferDescription("descriptors", (self.kpsize, 128), numpy.uint8, None),
                   BufferDescription("tmp", shape, numpy.float32, None),
                   BufferDescription("ori", shape, numpy.float32, None),
                   BufferDescription("DoGs", (par.Scales + 2, shape[0], shape[1]), numpy.float32, None),
                   BufferDescription("max_min", (self.red_size, 2), numpy.float32, None)  # temporary buffer for max/min reduction
                   ]
        if self.dtype != numpy.float32:
            if self.RGB:
                rgbshape = self.shape[0], self.shape[1], 3
                buffers.append(BufferDescription("raw", rgbshape, self.dtype, None))
            else:
                buffers.append(BufferDescription("raw", shape, self.dtype, None))
        for scale in range(par.Scales + 3):
            buffers.append(BufferDescription("scale_%i" % scale, shape, numpy.float32, None))

        self.allocate_buffers(buffers, use_array=True)

        self.cl_mem["255"].fill(255.0)
        ########################################################################
        # Allocate space for gaussian kernels
        ########################################################################
        curSigma = 1.0 if par.DoubleImSize else 0.5
        if self._init_sigma > curSigma:
            sigma = math.sqrt(self._init_sigma ** 2 - curSigma ** 2)
            self._init_gaussian(sigma)
        prevSigma = self._init_sigma

        for _ in range(par.Scales + 2):
            increase = prevSigma * math.sqrt(self.sigmaRatio ** 2 - 1.0)
            self._init_gaussian(increase)
            prevSigma *= self.sigmaRatio

    def _init_gaussian(self, sigma):
        """Create a buffer of the right size according to the width of the gaussian ...


        :param  sigma: width of the gaussian, the length of the function will be 8*sigma + 1

        Same calculation done on CPU
        x = numpy.arange(size) - (size - 1.0) / 2.0
        gaussian = numpy.exp(-(x / sigma) ** 2 / 2.0).astype(numpy.float32)
        gaussian /= gaussian.sum(dtype=numpy.float32)
        """
        pyopencl.enqueue_barrier(self.queue).wait()
        name = "gaussian_%s" % sigma
        size = kernel_size(sigma, True)
        wg_size = nextpower(size)

        logger.info("Allocating %s float for blur sigma: %s. wg=%s max_wg=%s", size, sigma, wg_size, self.block_size)
        wg1 = self.kernels_wg["gaussian"]
        if wg1 >= wg_size:
            gaussian_gpu = pyopencl.array.empty(self.queue, size, dtype=numpy.float32)
            pyopencl.enqueue_barrier(self.queue).wait()
            kernel = self.kernels.get_kernel("gaussian")
            shm1 = pyopencl.LocalMemory(4 * wg_size)
            shm2 = pyopencl.LocalMemory(4 * wg_size)
            evt = kernel(self.queue, (wg_size,), (wg_size,),
                         gaussian_gpu.data,
                         numpy.float32(sigma),  # const        float     sigma,
                         numpy.int32(size),  # const        int     SIZE
                         shm1, shm2)  # some shared memory
            pyopencl.enqueue_barrier(self.queue).wait()
            if self.profile:
                self.events.append(("gaussian %s" % sigma, evt))
        else:
            logger.info("Workgroup size error: gaussian wg: %s < max_work_group_size: %s",
                        wg1, self.block_size)
            # common bug on OSX when running on CPU
            x = numpy.arange(size) - (size - 1.0) / 2.0
            gaus = numpy.exp(-(x / sigma) ** 2 / 2.0).astype(numpy.float32)
            gaus /= gaus.sum(dtype=numpy.float32)
            gaussian_gpu = pyopencl.array.to_device(self.queue, gaus)

        self.cl_mem[name] = gaussian_gpu
        return gaussian_gpu

    def compile_kernels(self):
        """Call the OpenCL compiler

        TODO: use the parameters to define the compile-time constants and use 
        them all in kernels.
        """
        to_compile = ["sift",
                      "convolution",
                      "preprocess",
                      "algebra",
                      "image",
                      "gaussian",
                      "reductions",
                      "memset"]
        to_compile += list(self.kernels_max_wg_size.keys())
        compile_options = "-D WORKGROUP_SIZE=%s" % self.block_size
        try:
            OpenclProcessing.compile_kernels(self,
                                             [os.path.join("sift", kernel)
                                              for kernel in to_compile],
                                             compile_options=compile_options)
        except Exception as err:
            logger.error("error while compiling sift: %s", err)
        else:
            for kn in self.kernels.get_kernels():
                res = self.check_workgroup_size(kn)
                self.kernels_wg[kn] = min(res, self.block_size)

    def _free_kernels(self):
        """free all kernels
        """
        self.programs = {}

    def _calc_workgroups(self):
        """First try to guess the best workgroup size, then calculate all global worksize

        Nota:
        The workgroup size is limited by the device, some devices report wrong size.
        The workgroup size is limited to the 2**n below then image size (hence changes with octaves)
        The second dimension of the wg size should be large, the first small: i.e. (1,64)
        The processing size should be a multiple of  workgroup size.
        """
        device = self.ctx.devices[0]
        max_work_item_sizes = device.max_work_item_sizes
        if self.block_size:
            self.block_size = min(max_work_item_sizes[0], self.block_size)
        else:
            self.block_size = max_work_item_sizes[0]
        # MacOSX driver on CPU usually reports bad workgroup size: this is addressed in ocl
        self.block_size = min(self.block_size,
                              self.device.max_work_group_size)

        for k, v in self.kernels_max_wg_size.items():
            if isinstance(v, int):
                self.kernels_wg[k] = min(v, self.block_size)
            else:  # probably a list
                prod = numpy.prod(v)
                if prod <= self.block_size:
                    self.kernels_wg[k] = prod
                # else it is not possible to run this kernel.
                # If the kernel is not present in the dict, it should not be used.

        wg_float = min(self.block_size, numpy.sqrt(self.shape[0] * self.shape[1]))
        self.red_size = nextpower(wg_float)

        # we recalculate the shapes ...
        shape = self.shape
        min_size = 2 * par.BorderDist + 2
        while min(shape) > min_size:
            wg = (min(nextpower(shape[-1]), self.block_size), 1)
            self.wgsize.append(wg)
            self.procsize.append(calc_size(shape[-1::-1], wg))
            shape = tuple(i // 2 for i in shape)

    def keypoints(self, image, mask=None):
        """Calculates the keypoints of the image

        TODO: use a temporary list with events and use a single test at the end

        :param image: ndimage of 2D (or 3D if RGB)
        :param mask: TODO: implement a mask for sieving out the keypoints
        :return: vector of keypoint (1D numpy array)
        """
        # self.reset_timer()
        with self.sem:
            total_size = 0
            keypoints = []
            descriptors = []
            assert image.shape[:2] == self.shape
            assert image.dtype in [self.dtype, numpy.float32]
            # old versions of pyopencl do not check for data contiguity
            if not(isinstance(image, pyopencl.array.Array)) and not(image.flags["C_CONTIGUOUS"]):
                image = numpy.ascontiguousarray(image)
            t0 = time.time()

            if image.dtype == numpy.float32:
                if isinstance(image, pyopencl.array.Array):
                    evt = pyopencl.enqueue_copy(self.queue, self.cl_mem["scale_0"].data, image.data)
                else:
                    evt = pyopencl.enqueue_copy(self.queue, self.cl_mem["scale_0"].data, image)
                if self.profile:
                    self.events.append(("copy H->D", evt))
            elif self.dtype == numpy.float64:
                # A preprocessing kernel double_to_float exists, but is commented (RUNS ONLY ON GPU WITH FP64)
                # TODO: benchmark this kernel vs the current pure CPU format conversion with numpy.float32
                #       and uncomment it if it proves faster (dubious, because of data transfer bottleneck)
                evt = pyopencl.enqueue_copy(self.queue, self.cl_mem["scale_0"].data, image.astype(numpy.float32))
                if self.profile:
                    self.events.append(("copy H->D", evt))
            elif (len(image.shape) == 3) and (image.dtype == numpy.uint8) and (self.RGB):
                if isinstance(image, pyopencl.array.Array):
                    evt = pyopencl.enqueue_copy(self.queue, self.cl_mem["raw"].data, image.data)
                else:
                    evt = pyopencl.enqueue_copy(self.queue, self.cl_mem["raw"].data, image)
                if self.profile:
                    self.events.append(("copy H->D", evt))

                evt = self.kernels.get_kernel("rgb_to_float")(self.queue, self.procsize[0], self.wgsize[0],
                                                       self.cl_mem["raw"].data, self.cl_mem["scale_0"].data,
                                                       *self.scales[0])
                if self.profile:
                    self.events.append(("RGB -> float", evt))

            elif self.dtype in self.converter:
                program = self.kernels.get_kernel(self.converter[self.dtype])
                evt = pyopencl.enqueue_copy(self.queue, self.cl_mem["raw"].data, image)
                if self.profile:
                    self.events.append(("copy H->D", evt))
                evt = program(self.queue, self.procsize[0], self.wgsize[0],
                              self.cl_mem["raw"].data, self.cl_mem["scale_0"].data, *self.scales[0])
                if self.profile:
                    self.events.append(("convert -> float", evt))
            else:
                raise RuntimeError("invalid input format error (%s)" % (str(self.dtype)))

            wg1 = self.kernels_wg["max_min_global_stage1"]
            wg2 = self.kernels_wg["max_min_global_stage2"]
            if min(wg1, wg2) < self.red_size:
                # common bug on OSX when running on CPU
                logger.info("Unable to use MinMax Reduction: stage1 wg: %s; stage2 wg: %s < max_work_group_size: %s, expected: %s",
                            wg1, wg2, self.block_size, self.red_size)
                kernel = self.kernels.get_kernel("max_min_vec16")
                k = kernel(self.queue, (1,), (1,),
                               self.cl_mem["scale_0"].data,
                               numpy.int32(self.shape[0] * self.shape[1]),
                               self.cl_mem["max"].data,
                               self.cl_mem["min"].data)
                if self.profile:
                    self.events.append(("max_min_serial", k))
                # python implementation:
                # buffer_ = self.cl_mem["scale_0"].get()
                # self.cl_mem["max"].set(numpy.array([buffer_.max()], dtype=numpy.float32))
                # self.cl_mem["min"].set(numpy.array([buffer_.min()], dtype=numpy.float32))
            else:
                kernel1 = self.kernels.get_kernel("max_min_global_stage1")
                kernel2 = self.kernels.get_kernel("max_min_global_stage2")
                # logger.debug("self.red_size: %s", self.red_size)
                shm = pyopencl.LocalMemory(self.red_size * 2 * 4)
                k1 = kernel1(self.queue, (self.red_size * self.red_size,), (self.red_size,),
                             self.cl_mem["scale_0"].data,
                             self.cl_mem["max_min"].data,
                             numpy.int32(self.shape[0] * self.shape[1]),
                             shm)
                k2 = kernel2(self.queue, (self.red_size,), (self.red_size,),
                             self.cl_mem["max_min"].data,
                             self.cl_mem["max"].data,
                             self.cl_mem["min"].data,
                             shm)

                if self.profile:
                    self.events.append(("max_min_stage1", k1))
                    self.events.append(("max_min_stage2", k2))

            evt = self.kernels.get_kernel("normalizes")(self.queue, self.procsize[0], self.wgsize[0],
                                                        self.cl_mem["scale_0"].data,
                                                        self.cl_mem["min"].data,
                                                        self.cl_mem["max"].data,
                                                        self.cl_mem["255"].data,
                                                        *self.scales[0])
            if self.profile:
                self.events.append(("normalize", evt))

            curSigma = 1.0 if par.DoubleImSize else 0.5
            octave = 0
            if self._init_sigma > curSigma:
                logger.debug("Bluring image to achieve std: %f", self._init_sigma)
                sigma = math.sqrt(self._init_sigma ** 2 - curSigma ** 2)
                self._gaussian_convolution(self.cl_mem["scale_0"], self.cl_mem["scale_0"], sigma, 0)

            for octave in range(self.octave_max):
                kp, descriptor = self._one_octave(octave)
                logger.info("in octave %i found %i kp" % (octave, kp.shape[0]))

                if len(kp):
                    # sieve out coordinates with NaNs
                    mask = numpy.where(numpy.logical_not(numpy.isnan(kp.sum(axis=-1))))
                    keypoints.append(kp[mask])
                    descriptors.append(descriptor[mask])
                    total_size += len(mask[0])

            ########################################################################
            # Merge keypoints in central memory
            ########################################################################
            output = numpy.recarray(shape=(total_size,), dtype=self.dtype_kp)
            last = 0
            for ds, desc in zip(keypoints, descriptors):
                l = ds.shape[0]
                if l > 0:
                    output[last:last + l].x = ds[:, 0]
                    output[last:last + l].y = ds[:, 1]
                    output[last:last + l].scale = ds[:, 2]
                    output[last:last + l].angle = ds[:, 3]
                    output[last:last + l].desc = desc
                    last += l
            logger.info("Execution time: %.3fms" % (1000 * (time.time() - t0)))
        return output

    __call__ = keypoints

    def _gaussian_convolution(self, input_data, output_data, sigma, octave=0):
        """
        Calculate the gaussian convolution with precalculated kernels.

        :param input_data: pyopencl array with input
        :param output_data: pyopencl array with result
        :param sigma: width of the gaussian
        :param octave: related to the size on the input images

        * Uses a temporary buffer
        * Needs gaussian kernel to be available on device

        """
        temp_data = self.cl_mem["tmp"]
        gaussian = self.cl_mem["gaussian_%s" % sigma]
        k1 = self.kernels.get_kernel("horizontal_convolution")(self.queue, self.procsize[octave], self.wgsize[octave],
                                                               input_data.data, temp_data.data, gaussian.data, numpy.int32(gaussian.size),
                                                               *self.scales[octave])
        k2 = self.kernels.get_kernel("vertical_convolution")(self.queue, self.procsize[octave], self.wgsize[octave],
                                                             temp_data.data, output_data.data, gaussian.data, numpy.int32(gaussian.size),
                                                             *self.scales[octave])

        if self.profile:
            self.events += [("Blur sigma %s octave %s" % (sigma, octave), k1),
                            ("Blur sigma %s octave %s" % (sigma, octave), k2)]

    def _one_octave(self, octave):
        """
        Does all scales within an octave

        :param octave: number of the octave
        """
        prevSigma = self._init_sigma
        logger.info("Calculating octave %i" % octave)
        wgsize = (128,)  # (max(self.wgsize[octave]),) #TODO: optimize
        kpsize32 = numpy.int32(self.kpsize)
        self._reset_keypoints()
        octsize = numpy.int32(2 ** octave)
        last_start = numpy.int32(0)
        for scale in range(par.Scales + 2):
            sigma = prevSigma * math.sqrt(self.sigmaRatio ** 2 - 1.0)
            logger.info("Octave %i scale %s blur with sigma %s" % (octave, scale, sigma))

            ########################################################################
            # Calculate gaussian blur and DoG
            ########################################################################

            self._gaussian_convolution(self.cl_mem["scale_%i" % scale], self.cl_mem["scale_%i" % (scale + 1)], sigma, octave)
            prevSigma *= self.sigmaRatio
            evt = self.kernels.get_kernel("combine")(self.queue, self.procsize[octave], self.wgsize[octave],
                                                     self.cl_mem["scale_%i" % (scale + 1)].data, numpy.float32(-1.0),
                                                     self.cl_mem["scale_%i" % (scale)].data, numpy.float32(+1.0),
                                                     self.cl_mem["DoGs"].data, numpy.int32(scale),
                                                     *self.scales[octave])
            if self.profile:
                self.events.append(("DoG %s %s" % (octave, scale), evt))
        for scale in range(1, par.Scales + 1):
            evt = self.kernels.get_kernel("local_maxmin")(self.queue, self.procsize[octave], self.wgsize[octave],
                                                          self.cl_mem["DoGs"].data,  # __global float* DOGS,
                                                          self.cl_mem["Kp_1"].data,  # __global keypoint* output,
                                                          numpy.int32(par.BorderDist),  # int border_dist,
                                                          numpy.float32(par.PeakThresh),  # float peak_thresh,
                                                          octsize,  # int octsize,
                                                          numpy.float32(par.EdgeThresh1),  # float EdgeThresh0,
                                                          numpy.float32(par.EdgeThresh),  # float EdgeThresh,
                                                          self.cl_mem["cnt"].data,  # __global int* counter,
                                                          kpsize32,  # int nb_keypoints,
                                                          numpy.int32(scale),  # int scale,
                                                          *self.scales[octave])  # int width, int height)
            if self.profile:
                self.events.append(("local_maxmin %s %s" % (octave, scale), evt))
            procsize = calc_size((self.kpsize,), wgsize)
            cp_evt = pyopencl.enqueue_copy(self.queue, self.cnt, self.cl_mem["cnt"].data)
            # TODO: modify interp_keypoint so that it reads end_keypoint from GPU memory
            evt = self.kernels.get_kernel("interp_keypoint")(self.queue, procsize, wgsize,
                                                             self.cl_mem["DoGs"].data,  # __global float* DOGS,
                                                             self.cl_mem["Kp_1"].data,  # __global keypoint* keypoints,
                                                             last_start,  # int start_keypoint,
                                                             self.cnt[0],  # int end_keypoint,
                                                             numpy.float32(par.PeakThresh),  # float peak_thresh,
                                                             numpy.float32(self._init_sigma),  # float InitSigma,
                                                             *self.scales[octave])  # int width, int height)
            if self.profile:
                self.events += [("get cnt", cp_evt),
                                ("interp_keypoint %s %s" % (octave, scale), evt)
                                ]

            newcnt = self._compact(last_start)
            evt = self.kernels.get_kernel("compute_gradient_orientation")(self.queue, self.procsize[octave], self.wgsize[octave],
                                                                          self.cl_mem["scale_%s" % (scale)].data,  # __global float* igray,
                                                                          self.cl_mem["tmp"].data,  # __global float *grad,
                                                                          self.cl_mem["ori"].data,  # __global float *ori,
                                                                          *self.scales[octave])  # int width,int height
            if self.profile:
                self.events.append(("compute_gradient_orientation %s %s" % (octave, scale), evt))

#           Orientation assignement: 1D kernel, rather heavy kernel
            if newcnt and newcnt > last_start:  # launch kernel only if neededwgsize = (128,)

                if self.USE_CPU:
                    orientation_name = "orientation_cpu"
                    scales = self.scales[octave]
                else:
                    orientation_name = "orientation_gpu"
                    scales = list(self.scales[octave]) + \
                             [pyopencl.LocalMemory(36 * 4),
                              pyopencl.LocalMemory(128 * 4),
                              pyopencl.LocalMemory(128 * 4)]
                orientation = self.kernels.get_kernel(orientation_name)
                wg = self.kernels_max_wg_size[orientation_name]
                wgsize2 = (wg,)
                procsize = (int(newcnt * wg),)
                evt = orientation(self.queue, procsize, wgsize2,
                                  self.cl_mem["Kp_1"].data,  # __global keypoint* keypoints,
                                  self.cl_mem["tmp"].data,  # __global float* grad,
                                  self.cl_mem["ori"].data,  # __global float* ori,
                                  self.cl_mem["cnt"].data,  # __global int* counter,
                                  octsize,  # int octsize,
                                  numpy.float32(par.OriSigma),  # float OriSigma, //WARNING: (1.5), it is not "InitSigma (=1.6)"
                                  kpsize32,  # int max of nb_keypoints,
                                  numpy.int32(last_start),  # int keypoints_start,
                                  newcnt,  # int keypoints_end,
                                  *scales)  # int grad_width, int grad_height)
                # newcnt = self.cl_mem["cnt"].get()[0] #do not forget to update numbers of keypoints, modified above !
                evt_cp = pyopencl.enqueue_copy(self.queue, self.cnt, self.cl_mem["cnt"].data)
                newcnt = self.cnt[0]  # do not forget to update numbers of keypoints, modified above !

                for _ in range(3):
                    # up to 3 attempts
                    if self.USE_CPU or (self.LOW_END > 1):
                        logger.info("Computing descriptors with CPU optimized kernels")
                        descriptor_name = "descriptor_cpu"
                        wg = self.kernels_max_wg_size[descriptor_name][0]
                        wgsize2 = (wg,)
                        procsize2 = (int(newcnt * wg),)
                    else:
                        if self.LOW_END:
                            logger.info("Computing descriptors with older-GPU optimized kernels")
                            descriptor_name = "descriptor_gpu1"
                            wgsize2 = self.kernels_max_wg_size[descriptor_name]
                            procsize2 = (int(newcnt * wgsize2[0]), wgsize2[1], wgsize2[2])
#                             if self.kernels_wg[descriptor_name] < numpy.prod(wgsize2):
#                                 # will fail anyway:
#                                 self.LOW_END += 1
#                                 continue
                        else:
                            logger.info("Computing descriptors with newer-GPU optimized kernels")
                            descriptor_name = "descriptor_gpu2"
                            wgsize2 = self.kernels_max_wg_size[descriptor_name]
                            procsize2 = (int(newcnt * wgsize2[0]), wgsize2[1], wgsize2[2])
#                             if self.kernels_wg[descriptor_name] < numpy.prod(wgsize2):
#                                 # will fail anyway:
#                                 self.LOW_END += 1
#                                 continue
                    try:
                        descriptor = self.kernels.get_kernel(descriptor_name)
                        evt2 = descriptor(self.queue, procsize2, wgsize2,
                                          self.cl_mem["Kp_1"].data,  # __global keypoint* keypoints,
                                          self.cl_mem["descriptors"].data,  # ___global unsigned char *descriptors
                                          self.cl_mem["tmp"].data,  # __global float* grad,
                                          self.cl_mem["ori"].data,  # __global float* ori,
                                          octsize,  # int octsize,
                                          numpy.int32(last_start),  # int keypoints_start,
                                          self.cl_mem["cnt"].data,  # int* keypoints_end,
                                          *self.scales[octave])  # int grad_width, int grad_height)
                        evt2.wait()
                    except (pyopencl.RuntimeError, pyopencl._cl.LogicError) as error:
                        self.LOW_END += 1
                        logger.error("Descriptor failed with %s. Switching to lower_end mode" % error)
                        continue
                    else:
                        break
                if self.profile:
                    self.events += [("%s %s %s" % (orientation_name, octave, scale), evt),
                                    ("copy cnt D->H", evt_cp),
                                    ("%s %s %s" % (descriptor_name, octave, scale), evt2)]
            evt_cp = pyopencl.enqueue_copy(self.queue, self.cnt, self.cl_mem["cnt"].data)
            last_start = self.cnt[0]
            if self.profile:
                self.events.append(("copy cnt D->H", evt_cp))

        ########################################################################
        # Rescale all images to populate all octaves
        ########################################################################
        if octave < self.octave_max - 1:
            evt = self.kernels.get_kernel("shrink")(self.queue, self.procsize[octave + 1], self.wgsize[octave + 1],
                                                    self.cl_mem["scale_%i" % (par.Scales)].data,
                                                    self.cl_mem["scale_0"].data,
                                                    numpy.int32(2), numpy.int32(2),
                                                    self.scales[octave][0], self.scales[octave][1],
                                                    *self.scales[octave + 1])
            if self.profile:
                self.events.append(("shrink %s->%s" % (self.scales[octave], self.scales[octave + 1]), evt))
        results = numpy.empty((last_start, 4), dtype=numpy.float32)
        descriptors = numpy.empty((last_start, 128), dtype=numpy.uint8)
        if last_start:
            evt = pyopencl.enqueue_copy(self.queue, results, self.cl_mem["Kp_1"].data)
            evt2 = pyopencl.enqueue_copy(self.queue, descriptors, self.cl_mem["descriptors"].data)
            if self.profile:
                self.events += [("copy D->H", evt),
                                ("copy D->H", evt2)]
        return results, descriptors

    def _compact(self, start=numpy.int32(0)):
        """
        Compact the vector of keypoints starting from start

        :param start: start compacting at this adress. Before just copy
        :type  start: numpy.int32
        """
        wgsize = self.kernels_wg["compact"],
        cp0_evt = pyopencl.enqueue_copy(self.queue, self.cnt, self.cl_mem["cnt"].data)
        kp_counter = self.cnt[0]
        procsize = calc_size((self.kpsize,), wgsize)

        if kp_counter > 0.9 * self.kpsize:
            logger.warning("Keypoint counter overflow risk: counted %s / %s" % (kp_counter, self.kpsize))
        logger.info("Compact %s -> %s / %s" % (start, kp_counter, self.kpsize))
        self.cnt[0] = start
        cp1_evt = pyopencl.enqueue_copy(self.queue, self.cl_mem["cnt"].data, self.cnt)
        evt = self.kernels.get_kernel("compact")(self.queue, procsize, wgsize,
                                                 self.cl_mem["Kp_1"].data,  # __global keypoint* keypoints,
                                                 self.cl_mem["Kp_2"].data,  # __global keypoint* output,
                                                 self.cl_mem["cnt"].data,  # __global int* counter,
                                                 start,  # int start,
                                                 kp_counter)  # int nbkeypoints
        cp2_evt = pyopencl.enqueue_copy(self.queue, self.cnt, self.cl_mem["cnt"].data)
        # swap keypoints:
        self.cl_mem["Kp_1"], self.cl_mem["Kp_2"] = self.cl_mem["Kp_2"], self.cl_mem["Kp_1"]
        # memset buffer Kp_2
#        self.cl_mem["Kp_2"].fill(-1, self.queue)
        mem_evt = self.kernels.get_kernel("memset_float")(self.queue, calc_size((4 * self.kpsize,), wgsize), wgsize, self.cl_mem["Kp_2"].data, numpy.float32(-1), numpy.int32(4 * self.kpsize))
        if self.profile:
            self.events += [("copy cnt D->H", cp0_evt),
                            ("copy cnt H->D", cp1_evt),
                            ("compact", evt),
                            ("copy cnt D->H", cp2_evt),
                            ("memset 2", mem_evt)
                            ]
        return self.cnt[0]

    def _reset_keypoints(self):
        """
        Todo: implement directly in OpenCL instead of relying on pyOpenCL
        """
        wg_size = self.kernels_wg["memset_float"],
        evt1 = self.kernels.get_kernel("memset_float")(self.queue, calc_size((4 * self.kpsize,), wg_size), wg_size,
                                                       self.cl_mem["Kp_1"].data,
                                                       numpy.float32(-1),
                                                       numpy.int32(4 * self.kpsize))
#        evt2 = self.kernels.get_kernel("memset"].memset_float(self.queue, calc_size((4 * self.kpsize,), wg_size), wg_size, self.cl_mem["Kp_2"].data, numpy.float32(-1), numpy.int32(4 * self.kpsize))
        evt3 = self.kernels.get_kernel("memset_int")(self.queue, (1,), (1,),
                                                     self.cl_mem["cnt"].data,
                                                     numpy.int32(0),
                                                     numpy.int32(1))
        if self.profile:
            self.events += [("memset 1", evt1), ("memset cnt", evt3)]
#        self.cl_mem["Kp_1"].fill(-1, self.queue)
#        self.cl_mem["Kp_2"].fill(-1, self.queue)
#        self.cl_mem["cnt"].fill(0, self.queue)

    def count_kp(self, output):
        """
        Print the number of keypoint per octave
        """
        kpt = 0
        for octave, data in enumerate(output):
            if output.shape[0] > 0:
                ksum = (data[:, 1] != -1.0).sum()
                kpt += ksum
                print("octave %i kp count %i/%i size %s ratio:%s" % (octave, ksum, self.kpsize, self.scales[octave], 1000.0 * ksum / self.scales[octave][1] / self.scales[octave][0]))
        print("Found total %i guess %s pixels per keypoint" % (kpt, self.shape[0] * self.shape[1] / kpt))


def demo():
    # Prepare debugging
    import scipy.misc
    if hasattr(scipy.misc, "ascent"):
        img = scipy.misc.ascent()
    else:
        img = scipy.misc.lena()

    s = SiftPlan(template=img)
    print(s.keypoints(img))


if __name__ == "__main__":
    demo()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides a flow layout for QWidget: :class:`FlowLayout`.
"""

from __future__ import division

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "20/07/2018"


from .. import qt


class FlowLayout(qt.QLayout):
    """Layout widgets on (possibly) multiple lines in the available width.

    See Qt :class:`QLayout` for API documentation.

    Adapted from C++ `Qt FlowLayout example
    <http://doc.qt.io/qt-5/qtwidgets-layouts-flowlayout-example.html>`_

    :param QWidget parent: See :class:`QLayout`
    """

    def __init__(self, parent=None):
        super(FlowLayout, self).__init__(parent)
        self._items = []
        self._horizontalSpacing = -1
        self._verticalSpacing = -1

    def addItem(self, item):
        self._items.append(item)

    def count(self):
        return len(self._items)

    def itemAt(self, index):
        if 0 <= index < len(self._items):
            return self._items[index]
        else:
            return None

    def takeAt(self, index):
        if 0 <= index < len(self._items):
            return self._items.pop(index)
        else:
            return None

    def expandingDirections(self):
        return 0

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        return self._layout(qt.QRect(0, 0, width, 0), test=True)

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self._layout(rect)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = qt.QSize()
        for item in self._items:
            size = size.expandedTo(item.minimumSize())

        left, top, right, bottom = self.getContentsMargins()
        size += qt.QSize(left + right, top + bottom)
        return size

    def _layout(self, rect, test=False):
        left, top, right, bottom = self.getContentsMargins()
        effectiveRect = rect.adjusted(left, top, -right, -bottom)
        x, y = effectiveRect.x(), effectiveRect.y()
        lineHeight = 0

        for item in self._items:
            widget = item.widget()
            spaceX = self.horizontalSpacing()
            if spaceX == -1:
                spaceX = widget.style().layoutSpacing(
                    qt.QSizePolicy.PushButton,
                    qt.QSizePolicy.PushButton,
                    qt.Qt.Horizontal)
            spaceY = self.verticalSpacing()
            if spaceY == -1:
                spaceY = widget.style().layoutSpacing(
                    qt.QSizePolicy.PushButton,
                    qt.QSizePolicy.PushButton,
                    qt.Qt.Vertical)

            nextX = x + item.sizeHint().width() + spaceX
            if (nextX - spaceX) > effectiveRect.right() and lineHeight > 0:
                x = effectiveRect.x()
                y += lineHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0

            if not test:
                item.setGeometry(qt.QRect(qt.QPoint(x, y), item.sizeHint()))

            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())

        return y + lineHeight - rect.y() + bottom

    def setHorizontalSpacing(self, spacing):
        """Set the horizontal spacing between widgets laid out side by side

        :param int spacing:
        """
        self._horizontalSpacing = spacing
        self.update()

    def horizontalSpacing(self):
        """Returns the horizontal spacing between widgets laid out side by side

        :rtype: int
        """
        if self._horizontalSpacing >= 0:
            return self._horizontalSpacing
        else:
            return self._smartSpacing(qt.QStyle.PM_LayoutHorizontalSpacing)

    def setVerticalSpacing(self, spacing):
        """Set the vertical spacing between lines

        :param int spacing:
        """
        self._verticalSpacing = spacing
        self.update()

    def verticalSpacing(self):
        """Returns the vertical spacing between lines

        :rtype: int
        """
        if self._verticalSpacing >= 0:
            return self._verticalSpacing
        else:
            return self._smartSpacing(qt.QStyle.PM_LayoutVerticalSpacing)

    def _smartSpacing(self, pm):
        parent = self.parent()
        if parent is None:
            return -1
        if parent.isWidgetType():
            return parent.style().pixelMetric(pm, None, parent)
        else:
            return parent.spacing()



#!/usr/bin/env python
# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This script is an example to illustrate how to set range constraints on
plot axes.
"""

from silx.gui import qt
from silx.gui import plot
import numpy
import silx.test.utils


class ConstrainedViewPlot(qt.QMainWindow):

    def __init__(self):
        qt.QMainWindow.__init__(self)
        self.setWindowTitle("Plot with constrained axes")
        widget = qt.QWidget(self)
        self.setCentralWidget(widget)

        layout = qt.QGridLayout()
        widget.setLayout(layout)

        backend = "mpl"

        data = numpy.arange(100 * 100)
        data = (data % 100) / 5.0
        data = numpy.sin(data)
        data = silx.test.utils.add_gaussian_noise(data, mean=0.01)
        data.shape = 100, 100

        data1d = numpy.mean(data, axis=0)

        self.plot2d = plot.Plot2D(parent=widget, backend=backend)
        self.plot2d.setGraphTitle("A pixel can't be too big")
        self.plot2d.setInteractiveMode('pan')
        self.plot2d.addImage(data)
        self.plot2d.getXAxis().setRangeConstraints(minRange=10)
        self.plot2d.getYAxis().setRangeConstraints(minRange=10)

        self.plot2d2 = plot.Plot2D(parent=widget, backend=backend)
        self.plot2d2.setGraphTitle("The image can't be too small")
        self.plot2d2.setInteractiveMode('pan')
        self.plot2d2.addImage(data)
        self.plot2d2.getXAxis().setRangeConstraints(maxRange=200)
        self.plot2d2.getYAxis().setRangeConstraints(maxRange=200)

        self.plot1d = plot.Plot1D(parent=widget, backend=backend)
        self.plot1d.setGraphTitle("The curve is clamped into the view")
        self.plot1d.addCurve(x=numpy.arange(100), y=data1d, legend="mean")
        self.plot1d.getXAxis().setLimitsConstraints(minPos=0, maxPos=100)
        self.plot1d.getYAxis().setLimitsConstraints(minPos=data1d.min(), maxPos=data1d.max())

        self.plot1d2 = plot.Plot1D(parent=widget, backend=backend)
        self.plot1d2.setGraphTitle("Only clamp y-axis")
        self.plot1d2.setInteractiveMode('pan')
        self.plot1d2.addCurve(x=numpy.arange(100), y=data1d, legend="mean")
        self.plot1d2.getYAxis().setLimitsConstraints(minPos=data1d.min(), maxPos=data1d.max())

        layout.addWidget(self.plot2d, 0, 0)
        layout.addWidget(self.plot1d, 0, 1)
        layout.addWidget(self.plot2d2, 1, 0)
        layout.addWidget(self.plot1d2, 1, 1)


if __name__ == "__main__":
    app = qt.QApplication([])
    window = ConstrainedViewPlot()
    window.setVisible(True)
    app.exec_()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides 2D data and RGB(A) image item class.
"""

from __future__ import absolute_import

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "15/11/2017"

import numpy

from ..scene import primitives, utils
from .core import DataItem3D, ItemChangedType
from .mixins import ColormapMixIn, InterpolationMixIn
from ._pick import PickingResult


class _Image(DataItem3D, InterpolationMixIn):
    """Base class for images

    :param parent: The View widget this item belongs to.
    """

    def __init__(self, parent=None):
        DataItem3D.__init__(self, parent=parent)
        InterpolationMixIn.__init__(self)

    def _setPrimitive(self, primitive):
        InterpolationMixIn._setPrimitive(self, primitive)

    def getData(self, copy=True):
        raise NotImplementedError()

    def _pickFull(self, context):
        """Perform picking in this item at given widget position.

        :param PickContext context: Current picking context
        :return: Object holding the results or None
        :rtype: Union[None,PickingResult]
        """
        rayObject = context.getPickingSegment(frame=self._getScenePrimitive())
        if rayObject is None:
            return None

        points = utils.segmentPlaneIntersect(
            rayObject[0, :3],
            rayObject[1, :3],
            planeNorm=numpy.array((0., 0., 1.), dtype=numpy.float64),
            planePt=numpy.array((0., 0., 0.), dtype=numpy.float64))

        if len(points) == 1:  # Single intersection
            if points[0][0] < 0. or points[0][1] < 0.:
                return None  # Outside image
            row, column = int(points[0][1]), int(points[0][0])
            data = self.getData(copy=False)
            height, width = data.shape[:2]
            if row < height and column < width:
                return PickingResult(
                    self,
                    positions=[(points[0][0], points[0][1], 0.)],
                    indices=([row], [column]))
            else:
                return None  # Outside image
        else:  # Either no intersection or segment and image are coplanar
            return None


class ImageData(_Image, ColormapMixIn):
    """Description of a 2D image data.

    :param parent: The View widget this item belongs to.
    """

    def __init__(self, parent=None):
        _Image.__init__(self, parent=parent)
        ColormapMixIn.__init__(self)

        self._data = numpy.zeros((0, 0), dtype=numpy.float32)

        self._image = primitives.ImageData(self._data)
        self._getScenePrimitive().children.append(self._image)

        # Connect scene primitive to mix-in class
        ColormapMixIn._setSceneColormap(self, self._image.colormap)
        _Image._setPrimitive(self, self._image)

    def setData(self, data, copy=True):
        """Set the image data to display.

        The data will be casted to float32.

        :param numpy.ndarray data: The image data
        :param bool copy: True (default) to copy the data,
                          False to use as is (do not modify!).
        """
        self._image.setData(data, copy=copy)
        self._setColormappedData(self.getData(copy=False), copy=False)
        self._updated(ItemChangedType.DATA)

    def getData(self, copy=True):
        """Get the image data.

        :param bool copy:
            True (default) to get a copy,
            False to get internal representation (do not modify!).
        :rtype: numpy.ndarray
        :return: The image data
        """
        return self._image.getData(copy=copy)


class ImageRgba(_Image, InterpolationMixIn):
    """Description of a 2D data RGB(A) image.

    :param parent: The View widget this item belongs to.
    """

    def __init__(self, parent=None):
        _Image.__init__(self, parent=parent)
        InterpolationMixIn.__init__(self)

        self._data = numpy.zeros((0, 0, 3), dtype=numpy.float32)

        self._image = primitives.ImageRgba(self._data)
        self._getScenePrimitive().children.append(self._image)

        # Connect scene primitive to mix-in class
        _Image._setPrimitive(self, self._image)

    def setData(self, data, copy=True):
        """Set the RGB(A) image data to display.

        Supported array format: float32 in [0, 1], uint8.

        :param numpy.ndarray data:
            The RGBA image data as an array of shape (H, W, Channels)
        :param bool copy: True (default) to copy the data,
                          False to use as is (do not modify!).
        """
        self._image.setData(data, copy=copy)
        self._updated(ItemChangedType.DATA)

    def getData(self, copy=True):
        """Get the image data.

        :param bool copy:
            True (default) to get a copy,
            False to get internal representation (do not modify!).
        :rtype: numpy.ndarray
        :return: The image data
        """
        return self._image.getData(copy=copy)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2018-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ###########################################################################*/
"""Test SceneWidget picking feature"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "03/10/2018"


import unittest

import numpy

from silx.utils.testutils import ParametricTestCase
from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt

from silx.gui.plot3d.SceneWidget import SceneWidget, items


class TestSceneWidgetPicking(TestCaseQt, ParametricTestCase):
    """Tests SceneWidget picking feature"""

    def setUp(self):
        super(TestSceneWidgetPicking, self).setUp()
        self.widget = SceneWidget()
        self.widget.resize(300, 300)
        self.widget.show()
        # self.qWaitForWindowExposed(self.widget)

    def tearDown(self):
        self.qapp.processEvents()
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        del self.widget
        super(TestSceneWidgetPicking, self).tearDown()

    def _widgetCenter(self):
        """Returns widget center"""
        size = self.widget.size()
        return size.width() // 2, size.height() // 2

    def testPickImage(self):
        """Test picking of ImageData and ImageRgba items"""
        imageData = items.ImageData()
        imageData.setData(numpy.arange(100).reshape(10, 10))

        imageRgba = items.ImageRgba()
        imageRgba.setData(
            numpy.arange(300, dtype=numpy.uint8).reshape(10, 10, 3))

        for item in (imageData, imageRgba):
            with self.subTest(item=item.__class__.__name__):
                # Add item
                self.widget.clearItems()
                self.widget.addItem(item)
                self.widget.resetZoom('front')
                self.qapp.processEvents()

                # Picking on data (at widget center)
                picking = list(self.widget.pickItems(*self._widgetCenter()))

                self.assertEqual(len(picking), 1)
                self.assertIs(picking[0].getItem(), item)
                self.assertEqual(picking[0].getPositions('ndc').shape, (1, 3))
                data = picking[0].getData()
                self.assertEqual(len(data), 1)
                self.assertTrue(numpy.array_equal(
                    data,
                    item.getData()[picking[0].getIndices()]))

                # Picking outside data
                picking = list(self.widget.pickItems(1, 1))
                self.assertEqual(len(picking), 0)

    def testPickScatter(self):
        """Test picking of Scatter2D and Scatter3D items"""
        data = numpy.arange(100)

        scatter2d = items.Scatter2D()
        scatter2d.setData(x=data, y=data, value=data)

        scatter3d = items.Scatter3D()
        scatter3d.setData(x=data, y=data, z=data, value=data)

        for item in (scatter2d, scatter3d):
            with self.subTest(item=item.__class__.__name__):
                # Add item
                self.widget.clearItems()
                self.widget.addItem(item)
                self.widget.resetZoom('front')
                self.qapp.processEvents()

                # Picking on data (at widget center)
                picking = list(self.widget.pickItems(*self._widgetCenter()))

                self.assertEqual(len(picking), 1)
                self.assertIs(picking[0].getItem(), item)
                nbPos = len(picking[0].getPositions('ndc'))
                data = picking[0].getData()
                self.assertEqual(nbPos, len(data))
                self.assertTrue(numpy.array_equal(
                    data,
                    item.getValueData()[picking[0].getIndices()]))

                # Picking outside data
                picking = list(self.widget.pickItems(1, 1))
                self.assertEqual(len(picking), 0)

    def testPickVolume(self):
        """Test picking of volume CutPlane and Isosurface items"""
        for dtype in (numpy.float32, numpy.complex64):
            with self.subTest(dtype=dtype):
                refData = numpy.arange(10**3, dtype=dtype).reshape(10, 10, 10)
                volume = self.widget.addVolume(refData)
                if dtype == numpy.complex64:
                    volume.setComplexMode(volume.ComplexMode.REAL)
                    refData = numpy.real(refData)
                self.widget.resetZoom('front')

                cutplane = volume.getCutPlanes()[0]
                if dtype == numpy.complex64:
                    cutplane.setComplexMode(volume.ComplexMode.REAL)
                cutplane.getColormap().setVRange(0, 100)
                cutplane.setNormal((0, 0, 1))

                # Picking on data without anything displayed
                cutplane.setVisible(False)
                picking = list(self.widget.pickItems(*self._widgetCenter()))
                self.assertEqual(len(picking), 0)

                # Picking on data with the cut plane
                cutplane.setVisible(True)
                picking = list(self.widget.pickItems(*self._widgetCenter()))

                self.assertEqual(len(picking), 1)
                self.assertIs(picking[0].getItem(), cutplane)
                data = picking[0].getData()
                self.assertEqual(len(data), 1)
                self.assertEqual(picking[0].getPositions().shape, (1, 3))
                self.assertTrue(numpy.array_equal(
                    data,
                    refData[picking[0].getIndices()]))

                # Picking on data with an isosurface
                isosurface = volume.addIsosurface(
                    level=500, color=(1., 0., 0., .5))
                picking = list(self.widget.pickItems(*self._widgetCenter()))
                self.assertEqual(len(picking), 2)
                self.assertIs(picking[0].getItem(), cutplane)
                self.assertIs(picking[1].getItem(), isosurface)
                self.assertEqual(picking[1].getPositions().shape, (1, 3))
                data = picking[1].getData()
                self.assertEqual(len(data), 1)
                self.assertTrue(numpy.array_equal(
                    data,
                    refData[picking[1].getIndices()]))

                # Picking outside data
                picking = list(self.widget.pickItems(1, 1))
                self.assertEqual(len(picking), 0)

                self.widget.clearItems()

    def testPickMesh(self):
        """Test picking of Mesh items"""

        triangles = items.Mesh()
        triangles.setData(
            position=((0, 0, 0), (1, 0, 0), (1, 1, 0),
                      (0, 0, 0), (1, 1, 0), (0, 1, 0)),
            color=(1, 0, 0, 1),
            mode='triangles')
        triangleStrip = items.Mesh()
        triangleStrip.setData(
            position=(((1, 0, 0), (0, 0, 0), (1, 1, 0), (0, 1, 0))),
            color=(0, 1, 0, 1),
            mode='triangle_strip')
        triangleFan = items.Mesh()
        triangleFan.setData(
            position=((0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0)),
            color=(0, 0, 1, 1),
            mode='fan')

        for item in (triangles, triangleStrip, triangleFan):
            with self.subTest(mode=item.getDrawMode()):
                # Add item
                self.widget.clearItems()
                self.widget.addItem(item)
                self.widget.resetZoom('front')
                self.qapp.processEvents()

                # Picking on data (at widget center)
                picking = list(self.widget.pickItems(*self._widgetCenter()))

                self.assertEqual(len(picking), 1)
                self.assertIs(picking[0].getItem(), item)
                nbPos = len(picking[0].getPositions())
                data = picking[0].getData()
                self.assertEqual(nbPos, len(data))
                self.assertTrue(numpy.array_equal(
                    data,
                    item.getPositionData()[picking[0].getIndices()]))

                # Picking outside data
                picking = list(self.widget.pickItems(1, 1))
                self.assertEqual(len(picking), 0)

    def testPickMeshWithIndices(self):
        """Test picking of Mesh items defined by indices"""

        triangles = items.Mesh()
        triangles.setData(
            position=((0, 0, 0), (1, 0, 0), (0, 1, 0), (1, 1, 0)),
            color=(1, 0, 0, 1),
            indices=numpy.array(  # dummy triangles and square
                (0, 0, 1, 0, 1, 2, 1, 2, 3), dtype=numpy.uint8),
            mode='triangles')
        triangleStrip = items.Mesh()
        triangleStrip.setData(
            position=((0, 0, 0), (1, 0, 0), (0, 1, 0), (1, 1, 0)),
            color=(0, 1, 0, 1),
            indices=numpy.array(  # dummy triangles and square
                (1, 0, 0, 1, 2, 3), dtype=numpy.uint8),
            mode='triangle_strip')
        triangleFan = items.Mesh()
        triangleFan.setData(
            position=((0, 0, 0), (1, 0, 0), (0, 1, 0), (1, 1, 0)),
            color=(0, 0, 1, 1),
            indices=numpy.array(  # dummy triangle, square, dummy
                (1, 1, 0, 2, 3, 3), dtype=numpy.uint8),
            mode='fan')

        for item in (triangles, triangleStrip, triangleFan):
            with self.subTest(mode=item.getDrawMode()):
                # Add item
                self.widget.clearItems()
                self.widget.addItem(item)
                self.widget.resetZoom('front')
                self.qapp.processEvents()

                # Picking on data (at widget center)
                picking = list(self.widget.pickItems(*self._widgetCenter()))

                self.assertEqual(len(picking), 1)
                self.assertIs(picking[0].getItem(), item)
                nbPos = len(picking[0].getPositions())
                data = picking[0].getData()
                self.assertEqual(nbPos, len(data))
                self.assertTrue(numpy.array_equal(
                    data,
                    item.getPositionData()[picking[0].getIndices()]))

                # Picking outside data
                picking = list(self.widget.pickItems(1, 1))
                self.assertEqual(len(picking), 0)

    def testPickCylindricalMesh(self):
        """Test picking of Box, Cylinder and Hexagon items"""

        positions = numpy.array(((0., 0., 0.), (1., 1., 0.), (2., 2., 0.)))
        box = items.Box()
        box.setData(position=positions)
        cylinder = items.Cylinder()
        cylinder.setData(position=positions)
        hexagon = items.Hexagon()
        hexagon.setData(position=positions)

        for item in (box, cylinder, hexagon):
            with self.subTest(item=item.__class__.__name__):
                # Add item
                self.widget.clearItems()
                self.widget.addItem(item)
                self.widget.resetZoom('front')
                self.qapp.processEvents()

                # Picking on data (at widget center)
                picking = list(self.widget.pickItems(*self._widgetCenter()))

                self.assertEqual(len(picking), 1)
                self.assertIs(picking[0].getItem(), item)
                nbPos = len(picking[0].getPositions())
                data = picking[0].getData()
                print(item.__class__.__name__, [positions[1]], data)
                self.assertTrue(numpy.all(numpy.equal(positions[1], data)))
                self.assertEqual(nbPos, len(data))
                self.assertTrue(numpy.array_equal(
                    data,
                    item.getPosition()[picking[0].getIndices()]))

                # Picking outside data
                picking = list(self.widget.pickItems(1, 1))
                self.assertEqual(len(picking), 0)


def suite():
    testsuite = unittest.TestSuite()
    testsuite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(
            TestSceneWidgetPicking))
    return testsuite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Basic tests for MaskToolsWidget"""

__authors__ = ["T. Vincent", "P. Knobel"]
__license__ = "MIT"
__date__ = "17/01/2018"


import logging
import os.path
import unittest

import numpy

from silx.gui import qt
from silx.test.utils import temp_dir
from silx.utils.testutils import ParametricTestCase
from silx.gui.utils.testutils import getQToolButtonFromAction
from silx.gui.plot import PlotWindow, ScatterMaskToolsWidget
from .utils import PlotWidgetTestCase

import fabio


_logger = logging.getLogger(__name__)


class TestScatterMaskToolsWidget(PlotWidgetTestCase, ParametricTestCase):
    """Basic test for MaskToolsWidget"""

    def _createPlot(self):
        return PlotWindow()

    def setUp(self):
        super(TestScatterMaskToolsWidget, self).setUp()
        self.widget = ScatterMaskToolsWidget.ScatterMaskToolsDockWidget(
                plot=self.plot, name='TEST')
        self.plot.addDockWidget(qt.Qt.BottomDockWidgetArea, self.widget)

        self.maskWidget = self.widget.widget()

    def tearDown(self):
        del self.maskWidget
        del self.widget
        super(TestScatterMaskToolsWidget, self).tearDown()

    def testEmptyPlot(self):
        """Empty plot, display MaskToolsDockWidget, toggle multiple masks"""
        self.maskWidget.setMultipleMasks('single')
        self.qapp.processEvents()

        self.maskWidget.setMultipleMasks('exclusive')
        self.qapp.processEvents()

    def _drag(self):
        """Drag from plot center to offset position"""
        plot = self.plot.getWidgetHandle()
        xCenter, yCenter = plot.width() // 2, plot.height() // 2
        offset = min(plot.width(), plot.height()) // 10

        pos0 = xCenter, yCenter
        pos1 = xCenter + offset, yCenter + offset

        self.mouseMove(plot, pos=(0, 0))
        self.mouseMove(plot, pos=pos0)
        self.qapp.processEvents()
        self.mousePress(plot, qt.Qt.LeftButton, pos=pos0)
        self.qapp.processEvents()

        self.mouseMove(plot, pos=(pos0[0] + offset // 2, pos0[1] + offset // 2))
        self.mouseMove(plot, pos=pos1)
        self.qapp.processEvents()
        self.mouseRelease(plot, qt.Qt.LeftButton, pos=pos1)
        self.qapp.processEvents()
        self.mouseMove(plot, pos=(0, 0))

    def _drawPolygon(self):
        """Draw a star polygon in the plot"""
        plot = self.plot.getWidgetHandle()
        x, y = plot.width() // 2, plot.height() // 2
        offset = min(plot.width(), plot.height()) // 10

        star = [(x, y + offset),
                (x - offset, y - offset),
                (x + offset, y),
                (x - offset, y),
                (x + offset, y - offset),
                (x, y + offset)]  # Close polygon

        self.mouseMove(plot, pos=[0, 0])
        for pos in star:
            self.mouseMove(plot, pos=pos)
            self.qapp.processEvents()
            self.mousePress(plot, qt.Qt.LeftButton, pos=pos)
            self.qapp.processEvents()
            self.mouseRelease(plot, qt.Qt.LeftButton, pos=pos)
            self.qapp.processEvents()

    def _drawPencil(self):
        """Draw a star polygon in the plot"""
        plot = self.plot.getWidgetHandle()
        x, y = plot.width() // 2, plot.height() // 2
        offset = min(plot.width(), plot.height()) // 10

        star = [(x, y + offset),
                (x - offset, y - offset),
                (x + offset, y),
                (x - offset, y),
                (x + offset, y - offset)]

        self.mouseMove(plot, pos=[0, 0])
        self.mouseMove(plot, pos=star[0])
        self.mousePress(plot, qt.Qt.LeftButton, pos=star[0])
        for pos in star[1:]:
            self.mouseMove(plot, pos=pos)
        self.mouseRelease(
            plot, qt.Qt.LeftButton, pos=star[-1])

    def testWithAScatter(self):
        """Plot with a Scatter: test MaskToolsWidget interactions"""

        # Add and remove a scatter (this should enable/disable GUI + change mask)
        self.plot.addScatter(
                x=numpy.arange(256),
                y=numpy.arange(256),
                value=numpy.random.random(256),
                legend='test')
        self.plot._setActiveItem(kind="scatter", legend="test")
        self.qapp.processEvents()

        self.plot.remove('test', kind='scatter')
        self.qapp.processEvents()

        self.plot.addScatter(
                x=numpy.arange(1000),
                y=1000 * (numpy.arange(1000) % 20),
                value=numpy.random.random(1000),
                legend='test')
        self.plot._setActiveItem(kind="scatter", legend="test")
        self.plot.resetZoom()
        self.qapp.processEvents()

        # Test draw rectangle #
        toolButton = getQToolButtonFromAction(self.maskWidget.rectAction)
        self.assertIsNot(toolButton, None)
        self.mouseClick(toolButton, qt.Qt.LeftButton)

        # mask
        self.maskWidget.maskStateGroup.button(1).click()
        self.qapp.processEvents()
        self._drag()

        self.assertFalse(
            numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0)))

        # unmask same region
        self.maskWidget.maskStateGroup.button(0).click()
        self.qapp.processEvents()
        self._drag()
        self.assertTrue(
            numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0)))

        # Test draw polygon #
        toolButton = getQToolButtonFromAction(self.maskWidget.polygonAction)
        self.assertIsNot(toolButton, None)
        self.mouseClick(toolButton, qt.Qt.LeftButton)

        # mask
        self.maskWidget.maskStateGroup.button(1).click()
        self.qapp.processEvents()
        self._drawPolygon()
        self.assertFalse(
            numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0)))

        # unmask same region
        self.maskWidget.maskStateGroup.button(0).click()
        self.qapp.processEvents()
        self._drawPolygon()
        self.assertTrue(
            numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0)))

        # Test draw pencil #
        toolButton = getQToolButtonFromAction(self.maskWidget.pencilAction)
        self.assertIsNot(toolButton, None)
        self.mouseClick(toolButton, qt.Qt.LeftButton)

        self.maskWidget.pencilSpinBox.setValue(30)
        self.qapp.processEvents()

        # mask
        self.maskWidget.maskStateGroup.button(1).click()
        self.qapp.processEvents()
        self._drawPencil()
        self.assertFalse(
            numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0)))

        # unmask same region
        self.maskWidget.maskStateGroup.button(0).click()
        self.qapp.processEvents()
        self._drawPencil()
        self.assertTrue(
            numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0)))

        # Test no draw tool #
        toolButton = getQToolButtonFromAction(self.maskWidget.browseAction)
        self.assertIsNot(toolButton, None)
        self.mouseClick(toolButton, qt.Qt.LeftButton)

        self.plot.clear()

    def __loadSave(self, file_format):
        self.plot.addScatter(
                x=numpy.arange(256),
                y=25 * (numpy.arange(256) % 10),
                value=numpy.random.random(256),
                legend='test')
        self.plot._setActiveItem(kind="scatter", legend="test")
        self.plot.resetZoom()
        self.qapp.processEvents()

        # Draw a polygon mask
        toolButton = getQToolButtonFromAction(self.maskWidget.polygonAction)
        self.assertIsNot(toolButton, None)
        self.mouseClick(toolButton, qt.Qt.LeftButton)
        self._drawPolygon()

        ref_mask = self.maskWidget.getSelectionMask()
        self.assertFalse(numpy.all(numpy.equal(ref_mask, 0)))

        with temp_dir() as tmp:
            mask_filename = os.path.join(tmp, 'mask.' + file_format)
            self.maskWidget.save(mask_filename, file_format)

            self.maskWidget.resetSelectionMask()
            self.assertTrue(
                numpy.all(numpy.equal(self.maskWidget.getSelectionMask(), 0)))

            self.maskWidget.load(mask_filename)
            self.assertTrue(numpy.all(numpy.equal(
                self.maskWidget.getSelectionMask(), ref_mask)))

    def testLoadSaveNpy(self):
        self.__loadSave("npy")

    def testLoadSaveCsv(self):
        self.__loadSave("csv")

    def testSigMaskChangedEmitted(self):
        self.qapp.processEvents()
        self.plot.addScatter(
                x=numpy.arange(1000),
                y=1000 * (numpy.arange(1000) % 20),
                value=numpy.ones((1000,)),
                legend='test')
        self.plot._setActiveItem(kind="scatter", legend="test")
        self.plot.resetZoom()
        self.qapp.processEvents()

        self.plot.remove('test', kind='scatter')
        self.qapp.processEvents()

        self.plot.addScatter(
                x=numpy.arange(1000),
                y=1000 * (numpy.arange(1000) % 20),
                value=numpy.random.random(1000),
                legend='test')

        l = []

        def slot():
            l.append(1)

        self.maskWidget.sigMaskChanged.connect(slot)

        # rectangle mask
        toolButton = getQToolButtonFromAction(self.maskWidget.rectAction)
        self.assertIsNot(toolButton, None)
        self.mouseClick(toolButton, qt.Qt.LeftButton)
        self.maskWidget.maskStateGroup.button(1).click()
        self.qapp.processEvents()
        self._drag()

        self.assertGreater(len(l), 0)


def suite():
    test_suite = unittest.TestSuite()
    for TestClass in (TestScatterMaskToolsWidget,):
        test_suite.addTest(
            unittest.defaultTestLoader.loadTestsFromTestCase(TestClass))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""QWidget displaying a 2D image with histograms on its sides.

The :class:`ImageView` implements this widget, and
:class:`ImageViewMainWindow` provides a main window with additional toolbar
and status bar.

Basic usage of :class:`ImageView` is through the following methods:

- :meth:`ImageView.getColormap`, :meth:`ImageView.setColormap` to update the
  default colormap to use and update the currently displayed image.
- :meth:`ImageView.setImage` to update the displayed image.

For an example of use, see `imageview.py` in :ref:`sample-code`.
"""

from __future__ import division


__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "26/04/2018"


import logging
import numpy

import silx
from .. import qt

from . import items, PlotWindow, PlotWidget, actions
from ..colors import Colormap
from ..colors import cursorColorForColormap
from .tools import LimitsToolBar
from .Profile import ProfileToolBar
from ...utils.proxy import docstring

_logger = logging.getLogger(__name__)


# RadarView ###################################################################

class RadarView(qt.QGraphicsView):
    """Widget presenting a synthetic view of a 2D area and
    the current visible area.

    Coordinates are as in QGraphicsView:
    x goes from left to right and y goes from top to bottom.
    This widget preserves the aspect ratio of the areas.

    The 2D area and the visible area can be set with :meth:`setDataRect`
    and :meth:`setVisibleRect`.
    When the visible area has been dragged by the user, its new position
    is signaled by the *visibleRectDragged* signal.

    It is possible to invert the direction of the axes by using the
    :meth:`scale` method of QGraphicsView.
    """

    visibleRectDragged = qt.Signal(float, float, float, float)
    """Signals that the visible rectangle has been dragged.

    It provides: left, top, width, height in data coordinates.
    """

    _DATA_PEN = qt.QPen(qt.QColor('white'))
    _DATA_BRUSH = qt.QBrush(qt.QColor('light gray'))
    _VISIBLE_PEN = qt.QPen(qt.QColor('red'))
    _VISIBLE_PEN.setWidth(2)
    _VISIBLE_PEN.setCosmetic(True)
    _VISIBLE_BRUSH = qt.QBrush(qt.QColor(0, 0, 0, 0))
    _TOOLTIP = 'Radar View:\nRed contour: Visible area\nGray area: The image'

    _PIXMAP_SIZE = 256

    class _DraggableRectItem(qt.QGraphicsRectItem):
        """RectItem which signals its change through visibleRectDragged."""
        def __init__(self, *args, **kwargs):
            super(RadarView._DraggableRectItem, self).__init__(
                *args, **kwargs)

            self._previousCursor = None
            self.setFlag(qt.QGraphicsItem.ItemIsMovable)
            self.setFlag(qt.QGraphicsItem.ItemSendsGeometryChanges)
            self.setAcceptHoverEvents(True)
            self._ignoreChange = False
            self._constraint = 0, 0, 0, 0

        def setConstraintRect(self, left, top, width, height):
            """Set the constraint rectangle for dragging.

            The coordinates are in the _DraggableRectItem coordinate system.

            This constraint only applies to modification through interaction
            (i.e., this constraint is not applied to change through API).

            If the _DraggableRectItem is smaller than the constraint rectangle,
            the _DraggableRectItem remains within the constraint rectangle.
            If the _DraggableRectItem is wider than the constraint rectangle,
            the constraint rectangle remains within the _DraggableRectItem.
            """
            self._constraint = left, left + width, top, top + height

        def setPos(self, *args, **kwargs):
            """Overridden to ignore changes from API in itemChange."""
            self._ignoreChange = True
            super(RadarView._DraggableRectItem, self).setPos(*args, **kwargs)
            self._ignoreChange = False

        def moveBy(self, *args, **kwargs):
            """Overridden to ignore changes from API in itemChange."""
            self._ignoreChange = True
            super(RadarView._DraggableRectItem, self).moveBy(*args, **kwargs)
            self._ignoreChange = False

        def itemChange(self, change, value):
            """Callback called before applying changes to the item."""
            if (change == qt.QGraphicsItem.ItemPositionChange and
                    not self._ignoreChange):
                # Makes sure that the visible area is in the data
                # or that data is in the visible area if area is too wide
                x, y = value.x(), value.y()
                xMin, xMax, yMin, yMax = self._constraint

                if self.rect().width() <= (xMax - xMin):
                    if x < xMin:
                        value.setX(xMin)
                    elif x > xMax - self.rect().width():
                        value.setX(xMax - self.rect().width())
                else:
                    if x > xMin:
                        value.setX(xMin)
                    elif x < xMax - self.rect().width():
                        value.setX(xMax - self.rect().width())

                if self.rect().height() <= (yMax - yMin):
                    if y < yMin:
                        value.setY(yMin)
                    elif y > yMax - self.rect().height():
                        value.setY(yMax - self.rect().height())
                else:
                    if y > yMin:
                        value.setY(yMin)
                    elif y < yMax - self.rect().height():
                        value.setY(yMax - self.rect().height())

                if self.pos() != value:
                    # Notify change through signal
                    views = self.scene().views()
                    assert len(views) == 1
                    views[0].visibleRectDragged.emit(
                        value.x() + self.rect().left(),
                        value.y() + self.rect().top(),
                        self.rect().width(),
                        self.rect().height())

                return value

            return super(RadarView._DraggableRectItem, self).itemChange(
                change, value)

        def hoverEnterEvent(self, event):
            """Called when the mouse enters the rectangle area"""
            self._previousCursor = self.cursor()
            self.setCursor(qt.Qt.OpenHandCursor)

        def hoverLeaveEvent(self, event):
            """Called when the mouse leaves the rectangle area"""
            if self._previousCursor is not None:
                self.setCursor(self._previousCursor)
                self._previousCursor = None

    def __init__(self, parent=None):
        self._scene = qt.QGraphicsScene()
        self._dataRect = self._scene.addRect(0, 0, 1, 1,
                                             self._DATA_PEN,
                                             self._DATA_BRUSH)
        self._visibleRect = self._DraggableRectItem(0, 0, 1, 1)
        self._visibleRect.setPen(self._VISIBLE_PEN)
        self._visibleRect.setBrush(self._VISIBLE_BRUSH)
        self._scene.addItem(self._visibleRect)

        super(RadarView, self).__init__(self._scene, parent)
        self.setHorizontalScrollBarPolicy(qt.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(qt.Qt.ScrollBarAlwaysOff)
        self.setFocusPolicy(qt.Qt.NoFocus)
        self.setStyleSheet('border: 0px')
        self.setToolTip(self._TOOLTIP)

    def sizeHint(self):
        # """Overridden to avoid sizeHint to depend on content size."""
        return self.minimumSizeHint()

    def wheelEvent(self, event):
        # """Overridden to disable vertical scrolling with wheel."""
        event.ignore()

    def resizeEvent(self, event):
        # """Overridden to fit current content to new size."""
        self.fitInView(self._scene.itemsBoundingRect(), qt.Qt.KeepAspectRatio)
        super(RadarView, self).resizeEvent(event)

    def setDataRect(self, left, top, width, height):
        """Set the bounds of the data rectangular area.

        This sets the coordinate system.
        """
        self._dataRect.setRect(left, top, width, height)
        self._visibleRect.setConstraintRect(left, top, width, height)
        self.fitInView(self._scene.itemsBoundingRect(), qt.Qt.KeepAspectRatio)

    def setVisibleRect(self, left, top, width, height):
        """Set the visible rectangular area.

        The coordinates are relative to the data rect.
        """
        self._visibleRect.setRect(0, 0, width, height)
        self._visibleRect.setPos(left, top)
        self.fitInView(self._scene.itemsBoundingRect(), qt.Qt.KeepAspectRatio)


# ImageView ###################################################################

class ImageView(PlotWindow):
    """Display a single image with horizontal and vertical histograms.

    Use :meth:`setImage` to control the displayed image.
    This class also provides the :class:`silx.gui.plot.Plot` API.

    The :class:`ImageView` inherits from :class:`.PlotWindow` (which provides
    the toolbars) and also exposes :class:`.PlotWidget` API for further
    plot control (plot title, axes labels, aspect ratio, ...).

    :param parent: The parent of this widget or None.
    :param backend: The backend to use for the plot (default: matplotlib).
                    See :class:`.PlotWidget` for the list of supported backend.
    :type backend: str or :class:`BackendBase.BackendBase`
    """

    HISTOGRAMS_COLOR = 'blue'
    """Color to use for the side histograms."""

    HISTOGRAMS_HEIGHT = 200
    """Height in pixels of the side histograms."""

    IMAGE_MIN_SIZE = 200
    """Minimum size in pixels of the image area."""

    # Qt signals
    valueChanged = qt.Signal(float, float, float)
    """Signals that the data value under the cursor has changed.

    It provides: row, column, data value.

    When the cursor is over an histogram, either row or column is Nan
    and the provided data value is the histogram value
    (i.e., the sum along the corresponding row/column).
    Row and columns are either Nan or integer values.
    """

    def __init__(self, parent=None, backend=None):
        self._imageLegend = '__ImageView__image' + str(id(self))
        self._cache = None  # Store currently visible data information
        self._updatingLimits = False

        super(ImageView, self).__init__(parent=parent, backend=backend,
                                        resetzoom=True, autoScale=False,
                                        logScale=False, grid=False,
                                        curveStyle=False, colormap=True,
                                        aspectRatio=True, yInverted=True,
                                        copy=True, save=True, print_=True,
                                        control=False, position=False,
                                        roi=False, mask=True)
        if parent is None:
            self.setWindowTitle('ImageView')

        if silx.config.DEFAULT_PLOT_IMAGE_Y_AXIS_ORIENTATION == 'downward':
            self.getYAxis().setInverted(True)

        self._initWidgets(backend)

        self.profile = ProfileToolBar(plot=self)
        """"Profile tools attached to this plot.

        See :class:`silx.gui.plot.PlotTools.ProfileToolBar`
        """

        self.addToolBar(self.profile)

        # Sync PlotBackend and ImageView
        self._updateYAxisInverted()

    def _initWidgets(self, backend):
        """Set-up layout and plots."""
        self._histoHPlot = PlotWidget(backend=backend, parent=self)
        self._histoHPlot.getWidgetHandle().setMinimumHeight(
            self.HISTOGRAMS_HEIGHT)
        self._histoHPlot.getWidgetHandle().setMaximumHeight(
            self.HISTOGRAMS_HEIGHT)
        self._histoHPlot.setInteractiveMode('zoom')
        self._histoHPlot.sigPlotSignal.connect(self._histoHPlotCB)

        self.setPanWithArrowKeys(True)

        self.setInteractiveMode('zoom')  # Color set in setColormap
        self.sigPlotSignal.connect(self._imagePlotCB)
        self.getYAxis().sigInvertedChanged.connect(self._updateYAxisInverted)
        self.sigActiveImageChanged.connect(self._activeImageChangedSlot)

        self._histoVPlot = PlotWidget(backend=backend, parent=self)
        self._histoVPlot.getWidgetHandle().setMinimumWidth(
            self.HISTOGRAMS_HEIGHT)
        self._histoVPlot.getWidgetHandle().setMaximumWidth(
            self.HISTOGRAMS_HEIGHT)
        self._histoVPlot.setInteractiveMode('zoom')
        self._histoVPlot.sigPlotSignal.connect(self._histoVPlotCB)

        self._radarView = RadarView(parent=self)
        self._radarView.visibleRectDragged.connect(self._radarViewCB)

        self.__setCentralWidget()

    def __setCentralWidget(self):
        """Set central widget with all its content"""
        layout = qt.QGridLayout()
        layout.addWidget(self.getWidgetHandle(), 0, 0)
        layout.addWidget(self._histoVPlot.getWidgetHandle(), 0, 1)
        layout.addWidget(self._histoHPlot.getWidgetHandle(), 1, 0)
        layout.addWidget(self._radarView, 1, 1, 1, 2)
        layout.addWidget(self.getColorBarWidget(), 0, 2)

        layout.setColumnMinimumWidth(0, self.IMAGE_MIN_SIZE)
        layout.setColumnStretch(0, 1)
        layout.setColumnMinimumWidth(1, self.HISTOGRAMS_HEIGHT)
        layout.setColumnStretch(1, 0)

        layout.setRowMinimumHeight(0, self.IMAGE_MIN_SIZE)
        layout.setRowStretch(0, 1)
        layout.setRowMinimumHeight(1, self.HISTOGRAMS_HEIGHT)
        layout.setRowStretch(1, 0)

        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        centralWidget = qt.QWidget(self)
        centralWidget.setLayout(layout)
        self.setCentralWidget(centralWidget)

    @docstring(PlotWidget)
    def setBackend(self, backend):
        # Use PlotWidget here since we override PlotWindow behavior
        PlotWidget.setBackend(self, backend)
        self.__setCentralWidget()

    def _dirtyCache(self):
        self._cache = None

    def _updateHistograms(self):
        """Update histograms content using current active image."""
        activeImage = self.getActiveImage()
        if activeImage is not None:
            wasUpdatingLimits = self._updatingLimits
            self._updatingLimits = True

            data = activeImage.getData(copy=False)
            origin = activeImage.getOrigin()
            scale = activeImage.getScale()
            height, width = data.shape

            xMin, xMax = self.getXAxis().getLimits()
            yMin, yMax = self.getYAxis().getLimits()

            # Convert plot area limits to image coordinates
            # and work in image coordinates (i.e., in pixels)
            xMin = int((xMin - origin[0]) / scale[0])
            xMax = int((xMax - origin[0]) / scale[0])
            yMin = int((yMin - origin[1]) / scale[1])
            yMax = int((yMax - origin[1]) / scale[1])

            if (xMin < width and xMax >= 0 and
                    yMin < height and yMax >= 0):
                # The image is at least partly in the plot area
                # Get the visible bounds in image coords (i.e., in pixels)
                subsetXMin = 0 if xMin < 0 else xMin
                subsetXMax = (width if xMax >= width else xMax) + 1
                subsetYMin = 0 if yMin < 0 else yMin
                subsetYMax = (height if yMax >= height else yMax) + 1

                if (self._cache is None or
                        subsetXMin != self._cache['dataXMin'] or
                        subsetXMax != self._cache['dataXMax'] or
                        subsetYMin != self._cache['dataYMin'] or
                        subsetYMax != self._cache['dataYMax']):
                    # The visible area of data has changed, update histograms

                    # Rebuild histograms for visible area
                    visibleData = data[subsetYMin:subsetYMax,
                                       subsetXMin:subsetXMax]
                    histoHVisibleData = numpy.sum(visibleData, axis=0)
                    histoVVisibleData = numpy.sum(visibleData, axis=1)

                    self._cache = {
                        'dataXMin': subsetXMin,
                        'dataXMax': subsetXMax,
                        'dataYMin': subsetYMin,
                        'dataYMax': subsetYMax,

                        'histoH': histoHVisibleData,
                        'histoHMin': numpy.min(histoHVisibleData),
                        'histoHMax': numpy.max(histoHVisibleData),

                        'histoV': histoVVisibleData,
                        'histoVMin': numpy.min(histoVVisibleData),
                        'histoVMax': numpy.max(histoVVisibleData)
                    }

                    # Convert to histogram curve and update plots
                    # Taking into account origin and scale
                    coords = numpy.arange(2 * histoHVisibleData.size)
                    xCoords = (coords + 1) // 2 + subsetXMin
                    xCoords = origin[0] + scale[0] * xCoords
                    xData = numpy.take(histoHVisibleData, coords // 2)
                    self._histoHPlot.addCurve(xCoords, xData,
                                              xlabel='', ylabel='',
                                              replace=False,
                                              color=self.HISTOGRAMS_COLOR,
                                              linestyle='-',
                                              selectable=False)
                    vMin = self._cache['histoHMin']
                    vMax = self._cache['histoHMax']
                    vOffset = 0.1 * (vMax - vMin)
                    if vOffset == 0.:
                        vOffset = 1.
                    self._histoHPlot.getYAxis().setLimits(vMin - vOffset,
                                                          vMax + vOffset)

                    coords = numpy.arange(2 * histoVVisibleData.size)
                    yCoords = (coords + 1) // 2 + subsetYMin
                    yCoords = origin[1] + scale[1] * yCoords
                    yData = numpy.take(histoVVisibleData, coords // 2)
                    self._histoVPlot.addCurve(yData, yCoords,
                                              xlabel='', ylabel='',
                                              replace=False,
                                              color=self.HISTOGRAMS_COLOR,
                                              linestyle='-',
                                              selectable=False)
                    vMin = self._cache['histoVMin']
                    vMax = self._cache['histoVMax']
                    vOffset = 0.1 * (vMax - vMin)
                    if vOffset == 0.:
                        vOffset = 1.
                    self._histoVPlot.getXAxis().setLimits(vMin - vOffset,
                                                          vMax + vOffset)
            else:
                self._dirtyCache()
                self._histoHPlot.remove(kind='curve')
                self._histoVPlot.remove(kind='curve')

            self._updatingLimits = wasUpdatingLimits

    def _updateRadarView(self):
        """Update radar view visible area.

        Takes care of y coordinate conversion.
        """
        xMin, xMax = self.getXAxis().getLimits()
        yMin, yMax = self.getYAxis().getLimits()
        self._radarView.setVisibleRect(xMin, yMin, xMax - xMin, yMax - yMin)

    # Plots event listeners

    def _imagePlotCB(self, eventDict):
        """Callback for imageView plot events."""
        if eventDict['event'] == 'mouseMoved':
            activeImage = self.getActiveImage()
            if activeImage is not None:
                data = activeImage.getData(copy=False)
                height, width = data.shape

                # Get corresponding coordinate in image
                origin = activeImage.getOrigin()
                scale = activeImage.getScale()
                if (eventDict['x'] >= origin[0] and
                        eventDict['y'] >= origin[1]):
                    x = int((eventDict['x'] - origin[0]) / scale[0])
                    y = int((eventDict['y'] - origin[1]) / scale[1])

                    if x >= 0 and x < width and y >= 0 and y < height:
                        self.valueChanged.emit(float(x), float(y),
                                               data[y][x])

        elif eventDict['event'] == 'limitsChanged':
            self._updateHistogramsLimits()

    def _updateHistogramsLimits(self):
            # Do not handle histograms limitsChanged while
            # updating their limits from here.
            self._updatingLimits = True

            # Refresh histograms
            self._updateHistograms()

            xMin, xMax = self.getXAxis().getLimits()
            yMin, yMax = self.getYAxis().getLimits()

            # Set horizontal histo limits
            self._histoHPlot.getXAxis().setLimits(xMin, xMax)

            # Set vertical histo limits
            self._histoVPlot.getYAxis().setLimits(yMin, yMax)

            self._updateRadarView()

            self._updatingLimits = False

    def _histoHPlotCB(self, eventDict):
        """Callback for horizontal histogram plot events."""
        if eventDict['event'] == 'mouseMoved':
            if self._cache is not None:
                activeImage = self.getActiveImage()
                if activeImage is not None:
                    xOrigin = activeImage.getOrigin()[0]
                    xScale = activeImage.getScale()[0]

                    minValue = xOrigin + xScale * self._cache['dataXMin']

                    if eventDict['x'] >= minValue:
                        data = self._cache['histoH']
                        column = int((eventDict['x'] - minValue) / xScale)
                        if column >= 0 and column < data.shape[0]:
                            self.valueChanged.emit(
                                float('nan'),
                                float(column + self._cache['dataXMin']),
                                data[column])

        elif eventDict['event'] == 'limitsChanged':
            if (not self._updatingLimits and
                    eventDict['xdata'] != self.getXAxis().getLimits()):
                xMin, xMax = eventDict['xdata']
                self.getXAxis().setLimits(xMin, xMax)

    def _histoVPlotCB(self, eventDict):
        """Callback for vertical histogram plot events."""
        if eventDict['event'] == 'mouseMoved':
            if self._cache is not None:
                activeImage = self.getActiveImage()
                if activeImage is not None:
                    yOrigin = activeImage.getOrigin()[1]
                    yScale = activeImage.getScale()[1]

                    minValue = yOrigin + yScale * self._cache['dataYMin']

                    if eventDict['y'] >= minValue:
                        data = self._cache['histoV']
                        row = int((eventDict['y'] - minValue) / yScale)
                        if row >= 0 and row < data.shape[0]:
                            self.valueChanged.emit(
                                float(row + self._cache['dataYMin']),
                                float('nan'),
                                data[row])

        elif eventDict['event'] == 'limitsChanged':
            if (not self._updatingLimits and
                    eventDict['ydata'] != self.getYAxis().getLimits()):
                yMin, yMax = eventDict['ydata']
                self.getYAxis().setLimits(yMin, yMax)

    def _radarViewCB(self, left, top, width, height):
        """Slot for radar view visible rectangle changes."""
        if not self._updatingLimits:
            # Takes care of Y axis conversion
            self.setLimits(left, left + width, top, top + height)

    def _updateYAxisInverted(self, inverted=None):
        """Sync image, vertical histogram and radar view axis orientation."""
        if inverted is None:
            # Do not perform this when called from plot signal
            inverted = self.getYAxis().isInverted()

        self._histoVPlot.getYAxis().setInverted(inverted)

        # Use scale to invert radarView
        # RadarView default Y direction is from top to bottom
        # As opposed to Plot. So invert RadarView when Plot is NOT inverted.
        self._radarView.resetTransform()
        if not inverted:
            self._radarView.scale(1., -1.)
        self._updateRadarView()

        self._radarView.update()

    def _activeImageChangedSlot(self, previous, legend):
        """Handle Plot active image change.

        Resets side histograms cache
        """
        self._dirtyCache()
        self._updateHistograms()

    def getHistogram(self, axis):
        """Return the histogram and corresponding row or column extent.

        The returned value when an histogram is available is a dict with keys:

        - 'data': numpy array of the histogram values.
        - 'extent': (start, end) row or column index.
          end index is not included in the histogram.

        :param str axis: 'x' for horizontal, 'y' for vertical
        :return: The histogram and its extent as a dict or None.
        :rtype: dict
        """
        assert axis in ('x', 'y')
        if self._cache is None:
            return None
        else:
            if axis == 'x':
                return dict(
                    data=numpy.array(self._cache['histoH'], copy=True),
                    extent=(self._cache['dataXMin'], self._cache['dataXMax']))
            else:
                return dict(
                    data=numpy.array(self._cache['histoV'], copy=True),
                    extent=(self._cache['dataYMin'], self._cache['dataYMax']))

    def radarView(self):
        """Get the lower right radarView widget."""
        return self._radarView

    def setRadarView(self, radarView):
        """Change the lower right radarView widget.

        :param RadarView radarView: Widget subclassing RadarView to replace
                                    the lower right corner widget.
        """
        self._radarView.visibleRectDragged.disconnect(self._radarViewCB)
        self._radarView = radarView
        self._radarView.visibleRectDragged.connect(self._radarViewCB)
        self.centralWidget().layout().addWidget(self._radarView, 1, 1)

        self._updateYAxisInverted()

    # High-level API

    def getColormap(self):
        """Get the default colormap description.

        :return: A description of the current colormap.
                 See :meth:`setColormap` for details.
        :rtype: dict
        """
        return self.getDefaultColormap()

    def setColormap(self, colormap=None, normalization=None,
                    autoscale=None, vmin=None, vmax=None, colors=None):
        """Set the default colormap and update active image.

        Parameters that are not provided are taken from the current colormap.

        The colormap parameter can also be a dict with the following keys:

        - *name*: string. The colormap to use:
          'gray', 'reversed gray', 'temperature', 'red', 'green', 'blue'.
        - *normalization*: string. The mapping to use for the colormap:
          either 'linear' or 'log'.
        - *autoscale*: bool. Whether to use autoscale (True)
          or range provided by keys 'vmin' and 'vmax' (False).
        - *vmin*: float. The minimum value of the range to use if 'autoscale'
          is False.
        - *vmax*: float. The maximum value of the range to use if 'autoscale'
          is False.
        - *colors*: optional. Nx3 or Nx4 array of float in [0, 1] or uint8.
                    List of RGB or RGBA colors to use (only if name is None)

        :param colormap: Name of the colormap in
            'gray', 'reversed gray', 'temperature', 'red', 'green', 'blue'.
            Or the description of the colormap as a dict.
        :type colormap: dict or str.
        :param str normalization: Colormap mapping: 'linear' or 'log'.
        :param bool autoscale: Whether to use autoscale (True)
                               or [vmin, vmax] range (False).
        :param float vmin: The minimum value of the range to use if
                           'autoscale' is False.
        :param float vmax: The maximum value of the range to use if
                           'autoscale' is False.
        :param numpy.ndarray colors: Only used if name is None.
            Custom colormap colors as Nx3 or Nx4 RGB or RGBA arrays
        """
        cmap = self.getDefaultColormap()

        if isinstance(colormap, Colormap):
            # Replace colormap
            cmap = colormap

            self.setDefaultColormap(cmap)

            # Update active image colormap
            activeImage = self.getActiveImage()
            if isinstance(activeImage, items.ColormapMixIn):
                activeImage.setColormap(cmap)

        elif isinstance(colormap, dict):
            # Support colormap parameter as a dict
            assert normalization is None
            assert autoscale is None
            assert vmin is None
            assert vmax is None
            assert colors is None
            cmap._setFromDict(colormap)

        else:
            if colormap is not None:
                cmap.setName(colormap)
            if normalization is not None:
                cmap.setNormalization(normalization)
            if autoscale:
                cmap.setVRange(None, None)
            else:
                if vmin is not None:
                    cmap.setVMin(vmin)
                if vmax is not None:
                    cmap.setVMax(vmax)
            if colors is not None:
                cmap.setColormapLUT(colors)

        cursorColor = cursorColorForColormap(cmap.getName())
        self.setInteractiveMode('zoom', color=cursorColor)

    def setImage(self, image, origin=(0, 0), scale=(1., 1.),
                 copy=True, reset=True):
        """Set the image to display.

        :param image: A 2D array representing the image or None to empty plot.
        :type image: numpy.ndarray-like with 2 dimensions or None.
        :param origin: The (x, y) position of the origin of the image.
                       Default: (0, 0).
                       The origin is the lower left corner of the image when
                       the Y axis is not inverted.
        :type origin: Tuple of 2 floats: (origin x, origin y).
        :param scale: The scale factor to apply to the image on X and Y axes.
                      Default: (1, 1).
                      It is the size of a pixel in the coordinates of the axes.
                      Scales must be positive numbers.
        :type scale: Tuple of 2 floats: (scale x, scale y).
        :param bool copy: Whether to copy image data (default) or not.
        :param bool reset: Whether to reset zoom and ROI (default) or not.
        """
        self._dirtyCache()

        assert len(origin) == 2
        assert len(scale) == 2
        assert scale[0] > 0
        assert scale[1] > 0

        if image is None:
            self.remove(self._imageLegend, kind='image')
            return

        data = numpy.array(image, order='C', copy=copy)
        assert data.size != 0
        assert len(data.shape) == 2
        height, width = data.shape

        self.addImage(data,
                      legend=self._imageLegend,
                      origin=origin, scale=scale,
                      colormap=self.getColormap(),
                      resetzoom=False)
        self.setActiveImage(self._imageLegend)
        self._updateHistograms()

        self._radarView.setDataRect(origin[0],
                                    origin[1],
                                    width * scale[0],
                                    height * scale[1])

        if reset:
            self.resetZoom()
        else:
            self._updateHistogramsLimits()


# ImageViewMainWindow #########################################################

class ImageViewMainWindow(ImageView):
    """:class:`ImageView` with additional toolbars

    Adds extra toolbar and a status bar to :class:`ImageView`.
    """
    def __init__(self, parent=None, backend=None):
        self._dataInfo = None
        super(ImageViewMainWindow, self).__init__(parent, backend)
        self.setWindowFlags(qt.Qt.Window)

        self.getXAxis().setLabel('X')
        self.getYAxis().setLabel('Y')
        self.setGraphTitle('Image')

        # Add toolbars and status bar
        self.addToolBar(qt.Qt.BottomToolBarArea, LimitsToolBar(plot=self))

        self.statusBar()

        menu = self.menuBar().addMenu('File')
        menu.addAction(self.getOutputToolBar().getSaveAction())
        menu.addAction(self.getOutputToolBar().getPrintAction())
        menu.addSeparator()
        action = menu.addAction('Quit')
        action.triggered[bool].connect(qt.QApplication.instance().quit)

        menu = self.menuBar().addMenu('Edit')
        menu.addAction(self.getOutputToolBar().getCopyAction())
        menu.addSeparator()
        menu.addAction(self.getResetZoomAction())
        menu.addAction(self.getColormapAction())
        menu.addAction(actions.control.KeepAspectRatioAction(self, self))
        menu.addAction(actions.control.YAxisInvertedAction(self, self))

        menu = self.menuBar().addMenu('Profile')
        menu.addAction(self.profile.hLineAction)
        menu.addAction(self.profile.vLineAction)
        menu.addAction(self.profile.crossAction)
        menu.addAction(self.profile.lineAction)
        menu.addAction(self.profile.clearAction)

        # Connect to ImageView's signal
        self.valueChanged.connect(self._statusBarSlot)

    def _statusBarSlot(self, row, column, value):
        """Update status bar with coordinates/value from plots."""
        if numpy.isnan(row):
            msg = 'Column: %d, Sum: %g' % (int(column), value)
        elif numpy.isnan(column):
            msg = 'Row: %d, Sum: %g' % (int(row), value)
        else:
            msg = 'Position: (%d, %d), Value: %g' % (int(row), int(column),
                                                     value)
        if self._dataInfo is not None:
            msg = self._dataInfo + ', ' + msg

        self.statusBar().showMessage(msg)

    def setImage(self, image, *args, **kwargs):
        """Set the displayed image.

        See :meth:`ImageView.setImage` for details.
        """
        if hasattr(image, 'dtype') and hasattr(image, 'shape'):
            assert len(image.shape) == 2
            height, width = image.shape
            self._dataInfo = 'Data: %dx%d (%s)' % (width, height,
                                                   str(image.dtype))
            self.statusBar().showMessage(self._dataInfo)
        else:
            self._dataInfo = None

        # Set the new image in ImageView widget
        super(ImageViewMainWindow, self).setImage(image, *args, **kwargs)
        self.setStatusBar(None)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tests for the silx.gui.utils.glutils module."""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "15/01/2020"


import logging
import unittest
from silx.gui.utils.glutils import isOpenGLAvailable


_logger = logging.getLogger(__name__)


class TestIsOpenGLAvailable(unittest.TestCase):
    """Test isOpenGLAvailable"""

    def test(self):
        for version in ((2, 1), (2, 1), (1000, 1)):
            with self.subTest(version=version):
                result = isOpenGLAvailable(version=version)
                _logger.info("isOpenGLAvailable returned: %s", str(result))
                if version[0] == 1000:
                    self.assertFalse(result)
                if not result:
                    self.assertFalse(result.status)
                    self.assertTrue(len(result.error) > 0)
                else:
                    self.assertTrue(result.status)
                    self.assertTrue(len(result.error) == 0)


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(
        TestIsOpenGLAvailable))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
This module provides a :class:`QTreeView` dedicated to display plot3d models.

This module contains:
- :class:`ParamTreeView`: A QTreeView specific for plot3d parameters and scene.
- :class:`ParameterTreeDelegate`: The delegate for :class:`ParamTreeView`.
- A set of specific editors used by :class:`ParameterTreeDelegate`:
  :class:`FloatEditor`, :class:`Vector3DEditor`,
  :class:`Vector4DEditor`, :class:`IntSliderEditor`, :class:`BooleanEditor`
"""

from __future__ import absolute_import

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "05/12/2017"


import numbers
import sys

import six

from .. import qt
from ..widgets.FloatEdit import FloatEdit as _FloatEdit
from ._model import visitQAbstractItemModel


class FloatEditor(_FloatEdit):
    """Editor widget for float.

    :param parent: The widget's parent
    :param float value: The initial editor value
    """

    valueChanged = qt.Signal(float)
    """Signal emitted when the float value has changed"""

    def __init__(self, parent=None, value=None):
        super(FloatEditor, self).__init__(parent, value)
        self.setAlignment(qt.Qt.AlignLeft)
        self.editingFinished.connect(self._emit)

    def _emit(self):
        self.valueChanged.emit(self.value)

    value = qt.Property(float,
                        fget=_FloatEdit.value,
                        fset=_FloatEdit.setValue,
                        user=True,
                        notify=valueChanged)
    """Qt user property of the float value this widget edits"""


class Vector3DEditor(qt.QWidget):
    """Editor widget for QVector3D.

    :param parent: The widget's parent
    :param flags: The widgets's flags
    """

    valueChanged = qt.Signal(qt.QVector3D)
    """Signal emitted when the QVector3D value has changed"""

    def __init__(self, parent=None, flags=qt.Qt.Widget):
        super(Vector3DEditor, self).__init__(parent, flags)
        layout = qt.QHBoxLayout(self)
        # layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self._xEdit = _FloatEdit(parent=self, value=0.)
        self._xEdit.setAlignment(qt.Qt.AlignLeft)
        # self._xEdit.editingFinished.connect(self._emit)
        self._yEdit = _FloatEdit(parent=self, value=0.)
        self._yEdit.setAlignment(qt.Qt.AlignLeft)
        # self._yEdit.editingFinished.connect(self._emit)
        self._zEdit = _FloatEdit(parent=self, value=0.)
        self._zEdit.setAlignment(qt.Qt.AlignLeft)
        # self._zEdit.editingFinished.connect(self._emit)
        layout.addWidget(qt.QLabel('x:'))
        layout.addWidget(self._xEdit)
        layout.addWidget(qt.QLabel('y:'))
        layout.addWidget(self._yEdit)
        layout.addWidget(qt.QLabel('z:'))
        layout.addWidget(self._zEdit)
        layout.addStretch(1)

    def _emit(self):
        vector = self.value
        self.valueChanged.emit(vector)

    def getValue(self):
        """Returns the QVector3D value of this widget

        :rtype: QVector3D
        """
        return qt.QVector3D(
            self._xEdit.value(), self._yEdit.value(), self._zEdit.value())

    def setValue(self, value):
        """Set the QVector3D value

        :param QVector3D value: The new value
        """
        self._xEdit.setValue(value.x())
        self._yEdit.setValue(value.y())
        self._zEdit.setValue(value.z())
        self.valueChanged.emit(value)

    value = qt.Property(qt.QVector3D,
                        fget=getValue,
                        fset=setValue,
                        user=True,
                        notify=valueChanged)
    """Qt user property of the QVector3D value this widget edits"""


class Vector4DEditor(qt.QWidget):
    """Editor widget for QVector4D.

    :param parent: The widget's parent
    :param flags: The widgets's flags
    """

    valueChanged = qt.Signal(qt.QVector4D)
    """Signal emitted when the QVector4D value has changed"""

    def __init__(self, parent=None, flags=qt.Qt.Widget):
        super(Vector4DEditor, self).__init__(parent, flags)
        layout = qt.QHBoxLayout(self)
        # layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self._xEdit = _FloatEdit(parent=self, value=0.)
        self._xEdit.setAlignment(qt.Qt.AlignLeft)
        # self._xEdit.editingFinished.connect(self._emit)
        self._yEdit = _FloatEdit(parent=self, value=0.)
        self._yEdit.setAlignment(qt.Qt.AlignLeft)
        # self._yEdit.editingFinished.connect(self._emit)
        self._zEdit = _FloatEdit(parent=self, value=0.)
        self._zEdit.setAlignment(qt.Qt.AlignLeft)
        # self._zEdit.editingFinished.connect(self._emit)
        self._wEdit = _FloatEdit(parent=self, value=0.)
        self._wEdit.setAlignment(qt.Qt.AlignLeft)
        # self._wEdit.editingFinished.connect(self._emit)
        layout.addWidget(qt.QLabel('x:'))
        layout.addWidget(self._xEdit)
        layout.addWidget(qt.QLabel('y:'))
        layout.addWidget(self._yEdit)
        layout.addWidget(qt.QLabel('z:'))
        layout.addWidget(self._zEdit)
        layout.addWidget(qt.QLabel('w:'))
        layout.addWidget(self._wEdit)
        layout.addStretch(1)

    def _emit(self):
        vector = self.value
        self.valueChanged.emit(vector)

    def getValue(self):
        """Returns the QVector4D value of this widget

        :rtype: QVector4D
        """
        return qt.QVector4D(self._xEdit.value(), self._yEdit.value(),
                            self._zEdit.value(), self._wEdit.value())

    def setValue(self, value):
        """Set the QVector4D value

        :param QVector4D value: The new value
        """
        self._xEdit.setValue(value.x())
        self._yEdit.setValue(value.y())
        self._zEdit.setValue(value.z())
        self._wEdit.setValue(value.w())
        self.valueChanged.emit(value)

    value = qt.Property(qt.QVector4D,
                        fget=getValue,
                        fset=setValue,
                        user=True,
                        notify=valueChanged)
    """Qt user property of the QVector4D value this widget edits"""


class IntSliderEditor(qt.QSlider):
    """Slider editor widget for integer.

    Note: Tracking is disabled.

    :param parent: The widget's parent
    """

    def __init__(self, parent=None):
        super(IntSliderEditor, self).__init__(parent)
        self.setOrientation(qt.Qt.Horizontal)
        self.setSingleStep(1)
        self.setRange(0, 255)
        self.setValue(0)


class BooleanEditor(qt.QCheckBox):
    """Checkbox editor for bool.

    This is a QCheckBox with white background.

    :param parent: The widget's parent
    """

    def __init__(self, parent=None):
        super(BooleanEditor, self).__init__(parent)
        self.setStyleSheet("background: white;")


class ParameterTreeDelegate(qt.QStyledItemDelegate):
    """TreeView delegate specific to plot3d scene and object parameter tree.

    It provides additional editors.

    :param parent: Delegate's parent
    """

    EDITORS = {
        bool: BooleanEditor,
        float: FloatEditor,
        qt.QVector3D: Vector3DEditor,
        qt.QVector4D: Vector4DEditor,
    }
    """Specific editors for different type of data"""

    def __init__(self, parent=None):
        super(ParameterTreeDelegate, self).__init__(parent)

    def _fixVariant(self, data):
        """Fix PyQt4 zero vectors being stored as QPyNullVariant.

        :param data: Data retrieved from the model
        :return: Corresponding object
        """
        if qt.BINDING == 'PyQt4' and isinstance(data, qt.QPyNullVariant):
            typeName = data.typeName()
            if typeName == 'QVector3D':
                data = qt.QVector3D()
            elif typeName == 'QVector4D':
                data = qt.QVector4D()
        return data

    def paint(self, painter, option, index):
        """See :meth:`QStyledItemDelegate.paint`"""
        data = index.data(qt.Qt.DisplayRole)
        data = self._fixVariant(data)

        if isinstance(data, (qt.QVector3D, qt.QVector4D)):
            if isinstance(data, qt.QVector3D):
                text = '(x: %g; y: %g; z: %g)' % (data.x(), data.y(), data.z())
            elif isinstance(data, qt.QVector4D):
                text = '(%g; %g; %g; %g)' % (data.x(), data.y(), data.z(), data.w())
            else:
                text = ''

            painter.save()
            painter.setRenderHint(qt.QPainter.Antialiasing, True)

            # Select palette color group
            colorGroup = qt.QPalette.Inactive
            if option.state & qt.QStyle.State_Active:
                colorGroup = qt.QPalette.Active
            if not option.state & qt.QStyle.State_Enabled:
                colorGroup = qt.QPalette.Disabled

            # Draw background if selected
            if option.state & qt.QStyle.State_Selected:
                brush = option.palette.brush(colorGroup,
                                             qt.QPalette.Highlight)
                painter.fillRect(option.rect, brush)

            # Draw text
            if option.state & qt.QStyle.State_Selected:
                colorRole = qt.QPalette.HighlightedText
            else:
                colorRole = qt.QPalette.WindowText
            color = option.palette.color(colorGroup, colorRole)
            painter.setPen(qt.QPen(color))
            painter.drawText(option.rect, qt.Qt.AlignLeft, text)

            painter.restore()

            # The following commented code does the same as QPainter based code
            # but it does not work with PySide
            # self.initStyleOption(option, index)
            # option.text = text
            # widget = option.widget
            # style = qt.QApplication.style() if not widget else widget.style()
            # style.drawControl(qt.QStyle.CE_ItemViewItem, option, painter, widget)

        else:
            super(ParameterTreeDelegate, self).paint(painter, option, index)

    def _commit(self, *args):
        """Commit data to the model from editors"""
        sender = self.sender()
        self.commitData.emit(sender)

    def editorEvent(self, event, model, option, index):
        """See :meth:`QStyledItemDelegate.editorEvent`"""
        if (event.type() == qt.QEvent.MouseButtonPress and
                isinstance(index.data(qt.Qt.EditRole), qt.QColor)):
            initialColor = index.data(qt.Qt.EditRole)

            def callback(color):
                theModel = index.model()
                theModel.setData(index, color, qt.Qt.EditRole)

            dialog = qt.QColorDialog(self.parent())
            # dialog.setOption(qt.QColorDialog.ShowAlphaChannel, True)
            if sys.platform == 'darwin':
                # Use of native color dialog on macos might cause problems
                dialog.setOption(qt.QColorDialog.DontUseNativeDialog, True)
            dialog.setCurrentColor(initialColor)
            dialog.currentColorChanged.connect(callback)
            if dialog.exec_() == qt.QDialog.Rejected:
                # Reset color
                dialog.setCurrentColor(initialColor)

            return True
        else:
            return super(ParameterTreeDelegate, self).editorEvent(
                event, model, option, index)

    def createEditor(self, parent, option, index):
        """See :meth:`QStyledItemDelegate.createEditor`"""
        data = index.data(qt.Qt.EditRole)
        data = self._fixVariant(data)
        editorHint = index.data(qt.Qt.UserRole)

        if callable(editorHint):
            editor = editorHint()
            assert isinstance(editor, qt.QWidget)
            editor.setParent(parent)

        elif isinstance(data, numbers.Number) and editorHint is not None:
            # Use a slider
            editor = IntSliderEditor(parent)
            range_ = editorHint
            editor.setRange(*range_)
            editor.sliderReleased.connect(self._commit)

        elif isinstance(data, six.string_types) and editorHint is not None:
            # Use a combo box
            editor = qt.QComboBox(parent)
            if data not in editorHint:
                editor.addItem(data)
            editor.addItems(editorHint)

            index = editor.findText(data)
            editor.setCurrentIndex(index)

            editor.currentIndexChanged.connect(self._commit)

        else:
            # Handle overridden editors from Python
            # Mimic Qt C++ implementation
            for type_, editorClass in self.EDITORS.items():
                if isinstance(data, type_):
                    editor = editorClass(parent)
                    metaObject = editor.metaObject()
                    userProperty = metaObject.userProperty()
                    if userProperty.isValid() and userProperty.hasNotifySignal():
                        notifySignal = userProperty.notifySignal()
                        if hasattr(notifySignal, 'signature'):  # Qt4
                            signature = notifySignal.signature()
                        else:
                            signature = notifySignal.methodSignature()
                            if qt.BINDING == 'PySide2':
                                signature = signature.data()
                            else:
                                signature = bytes(signature)

                        if hasattr(signature, 'decode'):  # For PySide with python3
                            signature = signature.decode('ascii')
                        signalName = signature.split('(')[0]

                        signal = getattr(editor, signalName)
                        signal.connect(self._commit)
                    break

            else:  # Default handling for default types
                return super(ParameterTreeDelegate, self).createEditor(
                    parent, option, index)

        editor.setAutoFillBackground(True)
        return editor

    def setModelData(self, editor, model, index):
        """See :meth:`QStyledItemDelegate.setModelData`"""
        if isinstance(editor, tuple(self.EDITORS.values())):
            # Special handling of Python classes
            # Translation of QStyledItemDelegate::setModelData to Python
            # To make it work with Python QVariant wrapping/unwrapping
            name = editor.metaObject().userProperty().name()
            if not name:
                pass  # TODO handle the case of missing user property
            if name:
                if hasattr(editor, name):
                    value = getattr(editor, name)
                else:
                    value = editor.property(name)
                model.setData(index, value, qt.Qt.EditRole)

        else:
            super(ParameterTreeDelegate, self).setModelData(editor, model, index)


class ParamTreeView(qt.QTreeView):
    """QTreeView specific to handle plot3d scene and object parameters.

    It provides additional editors and specific creation of persistent editors.

    :param parent: The widget's parent.
    """

    def __init__(self, parent=None):
        super(ParamTreeView, self).__init__(parent)

        header = self.header()
        header.setMinimumSectionSize(128)  # For colormap pixmaps
        if hasattr(header, 'setSectionResizeMode'):  # Qt5
            header.setSectionResizeMode(qt.QHeaderView.ResizeToContents)
        else:  # Qt4
            header.setResizeMode(qt.QHeaderView.ResizeToContents)

        delegate = ParameterTreeDelegate()
        self.setItemDelegate(delegate)

        self.setSelectionBehavior(qt.QAbstractItemView.SelectRows)
        self.setSelectionMode(qt.QAbstractItemView.SingleSelection)

        self.expanded.connect(self._expanded)

        self.setEditTriggers(qt.QAbstractItemView.CurrentChanged |
                             qt.QAbstractItemView.DoubleClicked)

        self.__persistentEditors = set()

    def _openEditorForIndex(self, index):
        """Check if it has to open a persistent editor for a specific cell.

        :param QModelIndex index: The cell index
        """
        if index.flags() & qt.Qt.ItemIsEditable:
            data = index.data(qt.Qt.EditRole)
            editorHint = index.data(qt.Qt.UserRole)
            if (isinstance(data, bool) or
                    callable(editorHint) or
                    (isinstance(data, numbers.Number) and editorHint)):
                self.openPersistentEditor(index)
                self.__persistentEditors.add(index)

    def _openEditors(self, parent=qt.QModelIndex()):
        """Open persistent editors in a subtree starting at parent.

        :param QModelIndex parent: The root of the subtree to process.
        """
        model = self.model()
        if model is not None:
            for index in visitQAbstractItemModel(model, parent):
                self._openEditorForIndex(index)

    def setModel(self, model):
        """Set the model this TreeView is displaying

        :param QAbstractItemModel model:
        """
        super(ParamTreeView, self).setModel(model)
        self._openEditors()

    def rowsInserted(self, parent, start, end):
        """See :meth:`QTreeView.rowsInserted`"""
        super(ParamTreeView, self).rowsInserted(parent, start, end)
        model = self.model()
        if model is not None:
            for row in range(start, end+1):
                self._openEditorForIndex(model.index(row, 1, parent))
                self._openEditors(model.index(row, 0, parent))

    def _expanded(self, index):
        """Handle QTreeView expanded signal"""
        name = index.data(qt.Qt.DisplayRole)
        if name == 'Transform':
            rotateIndex = self.model().index(1, 0, index)
            self.setExpanded(rotateIndex, True)

    def dataChanged(self, topLeft, bottomRight, roles=()):
        """Handle model dataChanged signal eventually closing editors"""
        if roles:  # Qt 5
            super(ParamTreeView, self).dataChanged(topLeft, bottomRight, roles)
        else:  # Qt4 compatibility
            super(ParamTreeView, self).dataChanged(topLeft, bottomRight)
        if not roles or qt.Qt.UserRole in roles:  # Check editorHint update
            for row in range(topLeft.row(), bottomRight.row() + 1):
                for column in range(topLeft.column(), bottomRight.column() + 1):
                    index = topLeft.sibling(row, column)
                    if index.isValid():
                        if self._isPersistentEditorOpen(index):
                            self.closePersistentEditor(index)
                        self._openEditorForIndex(index)

    def _isPersistentEditorOpen(self, index):
        """Returns True if a persistent editor is opened for index

        :param QModelIndex index:
        :rtype: bool
        """
        return index in self.__persistentEditors

    def selectionCommand(self, index, event=None):
        """Filter out selection of not selectable items"""
        if index.flags() & qt.Qt.ItemIsSelectable:
            return super(ParamTreeView, self).selectionCommand(index, event)
        else:
            return qt.QItemSelectionModel.NoUpdate
